;; ##################################################################################
;; # ============================================================================== #
;; # MrEd Designer - code-generation.ss                                             #
;; # http://mreddesigner.lozi.org                                                   #
;; # Copyright (C) Lozi Jean-Pierre, 2004 - mailto:jean-pierre@lozi.org             #
;; # ============================================================================== #
;; #                                                                                #
;; # This program is free software; you can redistribute it and/or                  #
;; # modify it under the terms of the GNU General Public License                    #
;; # as published by the Free Software Foundation; either version 2                 #
;; # of the License, or (at your option) any later version.                         #
;; #                                                                                #
;; # This program is distributed in the hope that it will be useful,                #
;; # but WITHOUT ANY WARRANTY; without even the implied warranty of                 #
;; # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                  #
;; # GNU General Public License for more details.                                   #
;; #                                                                                #
;; # You should have received a copy of the GNU General Public License              #
;; # along with this program; if not, write to the Free Software                    #
;; # Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.    #
;; #                                                                                #
;; ##################################################################################

;; This function generates the code from the project-data-global
;; and writes it to p-out. The second optional argument is the file
;; name.
(define (generate-code p-out . file-name-list)

  ;; Returns the element id, given the element
  (define (get-id element)
    (if (assq 'id (cdr element))
        (cadr (assq 'id (cdr element)))
        #f))
  
  ;; Computes the given line into code
  (define (helper element parent)
    (define (sub-parse L)
      (cond ((null? L) L)
            ((< (length (car L)) 2) (sub-parse (cdr L)))
            ((eq? (caar L) 'pointer) (sub-parse (cdr L)))
            ((eq? (caar L) 'parent) (sub-parse (cdr L)))
            ((eq? (caar L) 'children) (sub-parse (cdr L)))
            ((eq? (caar L) 'collapsed) (sub-parse (cdr L)))
            ((eq? (caar L) 'selected) (sub-parse (cdr L)))
            ((eq? (caar L) 'id) (sub-parse (cdr L)))
            ((eq? (caar L) 'path) (sub-parse (cdr L)))
            (else (cons (car L) (sub-parse (cdr L))))))
    ;; Function body
    (list 'define (get-id element) (append (list 'new) 
                                           (list (string->symbol (string-append (symbol->string (car element)) "%")))
                                           (list (list 'parent parent))
                                               (sub-parse (cdr element)))))
  
  ;; The main loop
  (define (parse-list L parent)
    (if (not (null? L))
        ;; We parse the first element...
        (let ((current-element (car L)))
          
          ;; Call helper with the current element
          (if (not (eq? (car current-element) 'project))
              (begin
                ;; If we are writing to a file, we write comments
                (if (not (null? file-name-list))
                    (fprintf p-out ";; ~a definition.\n" (get-id current-element)))
                (pretty-print (helper current-element parent) p-out)
                ;; If we are parsing a frame we need the code to show it
                (if (eq? (car current-element) 'frame)
                    (begin
                      ;; If we are writing to a file, we write comments
                      (if (not (null? file-name-list))
                          (fprintf p-out "\n;; Show ~a..." (get-id current-element)))
                      ;; Back to the code...
                      (fprintf p-out "\n~s\n" `(send ,(get-id current-element) show #t))))
                (newline p-out)))
          
          ;; Do the current element have children?
          (if (assq 'children (cdar L))
              
              ;; Then parse the children!
              (parse-list (cdr (assq 'children (cdr current-element))) (get-id current-element)))
          
          ;; And then parse the brothers...
          (parse-list (cdr L) parent))))
  
  ;; First, we write the header, if written to a file.
  (if (not (null? file-name-list))
      (begin
        (fprintf p-out ";;=======================================================================\n")
        (fprintf p-out ";; Code generated with MrEd Designer - http://mreddesigner.lozi.org\n")
        (fprintf p-out ";; Any bug report, comments : mailto:jean-pierre@lozi.org \n")
        (fprintf p-out ";; File ~a \n" (let-values (((first second third) (split-path (car file-name-list)))) second))
        (fprintf p-out ";;=======================================================================\n\n")))
  
  ;; Let's call the main loop
  (parse-list project-data-global #f))

;; This callback is called when the user clicks the "Generate code..." button
;; from the toolbox frame.
(define (generate-code-callback)
  ;; This callback is called when the user clicks the "OK" button
  (define (ok-button-callback radio-box dialog)
    (let ((selection (send radio-box get-selection)))
      (send dialog show #f)
      (if (= selection 0)
        (generate-code-to-file)
        (generate-code-to-console))))
  ;; This callback is called when the user clicks the "Cancel" button
  (define (cancel-button-callback dialog)
    (send dialog show #f))
  ;; We build the whole dialog using a let*
 (let* ((dialog (new dialog% (label "Generate Code...") (parent #f) (x 300) (y 200)))
         ;; The main horizontal pane
         (main-horizontal-pane (new horizontal-pane% (parent dialog) (border 10) (alignment '(left top))))
         ;; The app icon
         (message (new message% (label 'app) (parent (new vertical-pane% (parent main-horizontal-pane)))))
         ;; The vertical pane
         (vertical-pane (new vertical-pane% (parent main-horizontal-pane)  (min-width 250)))
         ;; The first sub horizontal pane
         (sub-horizontal-pane-1 (new horizontal-pane% (parent vertical-pane) (border 0) (alignment '(center center))))
         ;; The message
         (message (new message% (label "Where do you want the code to be generated?") (parent sub-horizontal-pane-1)))
         ;; The radio buttons
         (radio-box (new radio-box% (label #f) (choices '("To a file" "To the console"))
                         (parent vertical-pane) (callback void) (vert-margin 10)))
         ;; The second sub horizontal pane
         (sub-horizontal-pane-2 (new horizontal-pane% (parent vertical-pane) (border 0) (alignment '(right top))))
         ;; The two buttons
         (okbtn (new button% (label "OK") (parent sub-horizontal-pane-2) 
                     (callback (lambda (button control-event) (ok-button-callback radio-box dialog))) (min-width 60)))
         (cancelbtn (new button% (label "Cancel") (parent sub-horizontal-pane-2) 
                         (callback (lambda (button control-event) (cancel-button-callback dialog))) (min-width 60))))
   ;; Center the dialog
   (send dialog center)
   ;; Show the modal dialog 
   (send dialog show #t)))

;; This function is called when the user chooses to generate the code to a file.
(define (generate-code-to-file)
  (let ((result (put-file "Please choose the output file..." #f #f #f "" '() '(("Scheme file (.ss)" "*.ss")("Scheme file (.scm)" "*.scm")))))
    (if result
        ;; The user did not choose the "Cancel" button.
        (call-with-output-file result 
          (lambda (p-out) (generate-code p-out result)) 'replace))))

;; This function is called when the user chooses to generate the code to the console.
(define (generate-code-to-console)
  (generate-code (current-output-port)))