;; ##################################################################################
;; # ============================================================================== #
;; # MrEd Designer - files.ss                                                       #
;; # http://mreddesigner.lozi.org                                                   #
;; # Copyright (C) Lozi Jean-Pierre, 2004 - mailto:jean-pierre@lozi.org             #
;; # ============================================================================== #
;; #                                                                                #
;; # This program is free software; you can redistribute it and/or                  #
;; # modify it under the terms of the GNU General Public License                    #
;; # as published by the Free Software Foundation; either version 2                 #
;; # of the License, or (at your option) any later version.                         #
;; #                                                                                #
;; # This program is distributed in the hope that it will be useful,                #
;; # but WITHOUT ANY WARRANTY; without even the implied warranty of                 #
;; # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                  #
;; # GNU General Public License for more details.                                   #
;; #                                                                                #
;; # You should have received a copy of the GNU General Public License              #
;; # along with this program; if not, write to the Free Software                    #
;; # Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.    #
;; #                                                                                #
;; ##################################################################################

;; Function called when the user chooses "File->New"
(define (files-new)
  ; check whether to save a modified project
  (if (not no-change-since-last-save)
    (let
      ((save? (message-box "Warning" 
                           (string-append "Do you want to save the current project\n"
                                          "before creating a new one?")
                           #f '(yes-no caution))))
      (if (eq? 'yes save?)
        (files-save))))
  ; create a new project and update the windows
  (preview-update-preview project-data-global '((project (label "Untitled Project")
                                                         (path "")
                                                         (selected #f)
                                                         (collapsed #f))))
  ; update the hierachy widget
  (send hierarchy-widget set-data project-data-global)
  ; update the buttons for the hierarchy
  (update-hierarchy-buttons (get-selected-element project-data-global))
  ; update the property windows
  (send properties-widget empty)
  (send properties-widget update-in)
  (set! no-change-since-last-save #t))

;; Function called when the user chooses "File->Open"
(define (files-open)
  (let 
    ((result (if (not no-change-since-last-save)
               (message-box/custom "Warning" 
                                   (string-append 
                                       "Do you want to save the current project\n"
                                       "before opening a new one?")
                                   "Yes" "No" "Cancel" #f 
                                   '(default=1 number-order caution) 
                                   3)
               2)))
    (if (= 1 result) (files-save))
    (if (not (= 3 result))
      (let ((result-file (get-file "Open..." #f #f #f "*.med" '() 
                                   '(("MrEd Designer Project (.med)" "*.med"))) ))
        (if result-file
          (let ((result (path->string result-file)))
            (if (file-exists? result)
              (call-with-input-file result
                (lambda (p-in)
                  (if (equal? (read p-in) '(*med-file*))
                    (begin
                      (let 
                        ((new-data (read p-in))
                         (old-data project-data-global)
                         (unselect-func (lambda (x) (set-property-select x #f)))
                        )
                        ; unselect everything
                        (set! project-data-global 
                              (apply-function-on-data-list new-data unselect-func))
                        ;; We have to update the label and the path
                        ;; it also notifies the hierarchy widgets
                        (project-data-set-path-label result)
                        ;; update previews
                        (preview-update-preview old-data project-data-global)
                        ; notify the hierarchy window
                        (send hierarchy-widget set-data project-data-global)
                        ; update the hierarchy buttons
                        (update-hierarchy-buttons (get-selected-element project-data-global))
                        ; notify properties window
                        (send properties-widget empty)
                        (send properties-widget update-in)
                        (set! no-change-since-last-save #t)))
                    (message-box "Error" "The file format is not valid." #f '(ok stop)))))
              (message-box "Error" "The file does not exist." #f '(ok stop)))))))))

  
;; Function called when the user chooses "File->Save As..."
(define (files-save-as)
  ;; Removes unnecessary tags from project-data-global
  ;; pointer : removed
  ;; path : reset to ""
  (define (parse-line L)
    (cond ((null? L) '())
          ((symbol? (car L)) (cons (car L) (parse-line (cdr L))))
          ((eq? (caar L) 'pointer) (parse-line (cdr L)))
          ((eq? (caar L) 'path) (cons (list 'path "") (parse-line (cdr L))))
          (else (cons (car L) (parse-line (cdr L))))))
  ;; Function body
  (let ((result-file (put-file "Save as..." #f #f #f "*.med" '() 
                               '(("MrEd Designer Project (.med)" "*.med"))) ))
    (if result-file
      (let ((result (path->string result-file)))
        (call-with-output-file result 
          (lambda (p-out)
            (fprintf p-out "~s\n" '(*med-file*))
            (pretty-print (apply-function-on-data-list project-data-global 
                                                       parse-line) p-out))
          'replace)
        ;; We update the project label and path
        (project-data-set-path-label result)
        ; notify the hierarchy window
        (send hierarchy-widget set-data project-data-global)
        ; update the hierarchy buttons
        (update-hierarchy-buttons (get-selected-element project-data-global))
        ; notify properties window
        (send properties-widget update-in)
        (set! no-change-since-last-save #t)
      )
    )
    result-file))

;; Function called when the user chooses "File->Save"
(define (files-save)
  ;; Removes unnecessary tags from project-data-global
  (define (parse-line L)
    (cond ((null? L) '())
          ((symbol? (car L)) (cons (car L) (parse-line (cdr L))))
          ((eq? (caar L) 'pointer) (parse-line (cdr L)))
          ((eq? (caar L) 'path) (cons (list 'path "") (parse-line (cdr L))))
          (else (cons (car L) (parse-line (cdr L))))))
  ;; Function body
  (let ((path (project-data-get-path)))
    (if (not (string=? path ""))
        (if (file-exists? path)
            (begin
              (call-with-output-file path
                (lambda (p-out)
                  (fprintf p-out "~s\n" '(*med-file*))
                  (pretty-print (apply-function-on-data-list project-data-global 
                                                             parse-line) p-out))
                'replace)
              ;; Then we update the no-change-since-last-save variable...
              (set! no-change-since-last-save #t)
              path
            )
            (begin
              (message-box "Error" 
                           (string-append
                             "Unable to save the project because the file does not exist anymore.\n"
                             "Please check it has not been either deleted, moved or renamed.")
                           #f '(ok stop))
              (files-save-as)))
        (files-save-as))))


(define (files-quit-final)
  (define (hide-if-frame line)
    (if (eq? (car line) 'frame)
      (let ((window (get-property-value line 'pointer)))
        (if (send window is-shown?)
          (send window show #f)))))
  ;; we hide all previews
  (apply-procedure-on-data-list project-data-global hide-if-frame)
  ;; hide the hierarchy window and properties window
  (if (send hierarchy-frame is-shown?)
    (send hierarchy-frame  show #f)
  )
  (if (send properties-frame is-shown?)
    (send properties-frame show #f)
  )
  (if (send toolbox-frame is-shown?)
    (send toolbox-frame show #f)
  )
)

;; Function called when the user chooses "File->Quit"
(define (files-quit cancel)
  (let 
    ((result (if (not no-change-since-last-save)
               (if cancel
                 (message-box/custom "Warning" 
                                     "Would you like to save changes before quitting?"
                                     "Yes" "No" "Cancel" #f '(default=1 number-order caution) 3)
                 (message-box/custom "Warning" 
                                     "Would you like to save changes before quitting?"
                                     "Yes" "No" #f #f '(default=1 number-order caution) 3))
               2)))
    (cond
      ((= 1 result) ; Yes
       (if cancel
         (if (files-save)
           (files-quit-final)
           (message-box "Warning" 
                        (string-append
                             "Unable to save the project.\n"
                             "Program is not quitting.")
                           #f '(ok))
         )
         (begin
           (files-save)
           (files-quit-final)
         )
        )
       )
      ((= 2 result)
       (files-quit-final)))))


