;; ##################################################################################
;; # ============================================================================== #
;; # MrEd Designer - help.ss                                                        #
;; # http://mreddesigner.lozi.org                                                   #
;; # Copyright (C) Lozi Jean-Pierre, 2004 - mailto:jean-pierre@lozi.org             #
;; # ============================================================================== #
;; #                                                                                #
;; # This program is free software; you can redistribute it and/or                  #
;; # modify it under the terms of the GNU General Public License                    #
;; # as published by the Free Software Foundation; either version 2                 #
;; # of the License, or (at your option) any later version.                         #
;; #                                                                                #
;; # This program is distributed in the hope that it will be useful,                #
;; # but WITHOUT ANY WARRANTY; without even the implied warranty of                 #
;; # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                  #
;; # GNU General Public License for more details.                                   #
;; #                                                                                #
;; # You should have received a copy of the GNU General Public License              #
;; # along with this program; if not, write to the Free Software                    #
;; # Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.    #
;; #                                                                                #
;; ##################################################################################

;; This function is called when the user chooses "Online Help..." in the help menu
(define (help-online-help)
  (send-url "http://mreddesigner.lozi.org/support.php"))

;; This function is called when the user chooses "MrEd Help (from PLT)..." in the help menu
(define (help-mred-help)
  (send-url "http://download.plt-scheme.org/doc/206p1/html/mred/mred.html"))

;; This function is called when the user chooses "About MrEd Designer..." in the help menu
(define (help-about-dialog)  
  ;; Let's build the whole dialog using a let*
  (let*  ((dialog (new dialog% (label "About MrEd Designer...") (parent #f) (width 510) (height 259)))
          ;; The main vertical pane...
          (vertical-pane (new vertical-pane% (parent dialog) (border 5)))
          ;; We call the derived canvas class...
          (canvas (new about-canvas% (parent vertical-pane) (style '(border)) (min-width 510) (min-height 259)))
          ;; A very, very summed up information about the license...
          (message1 (new message% (label "This software is distributed under the terms of the General Public License (GPL),") (parent vertical-pane)))
          (message2 (new message% (label "either version 2 of the license, or (at your option) any later version.") (parent vertical-pane)))
          ;; The buttons' pane...
          (horizontal-pane (new horizontal-panel% (parent vertical-pane) (alignment '(center center))))
          ;; The 3 buttons...
          (button (new button% (label "Contact...") (min-width 166)(parent horizontal-pane)
                       (callback (lambda (button control-event)
                                   (send-url "mailto:jean-pierre@lozi.org")))))
          (button (new button% (label "Website...") (min-width 166)(parent horizontal-pane)
                       (callback (lambda (button control-event)
                                   (send-url "http://mreddesigner.lozi.org")))))
          (button (new button% (label "Close") (min-width 166)(parent horizontal-pane)
                       (callback (lambda (button control-event)
                                   (send dialog show #f))))))
    ;; The main function body
    (send dialog center)
    (send dialog show #t)))

;; The picture in the about dialog is a derived canvas, which undelines the links when the mouse enters in their
;; respective areas and which launches the appropriate link targets when the user clicks on the them.
(define about-canvas%
  ;; This class is inherited from a canvas.
  (class canvas%
    ;; Here are some drawing constants...
    (define k-black (make-object color% 0 0 0))
    (define k-underline-brush (make-object brush% (make-object color% 0 0 255) 'solid))
    (define k-invisible-pen (make-object pen%  k-black 0 'transparent))
    (define k-arrow-cursor (make-object cursor% 'arrow))
    ;; We cannot use the default hand cursor because under MacOS X, for example, it is an open hand...
    ;; Here is something really strange I found out : under MacOS X the image and mask have to be swapped,
    ;; I reckon this is a MrEd Bug.
    (define k-hand-cursor (if (eq? (system-type) 'macosx)
                              (make-object cursor% 
                                (make-object bitmap% 
                                  (build-path "images" "hand-cursor-mask.xbm")
                                  'xbm (make-object color% 255 255 255))
                                (make-object bitmap% 
                                  (build-path "images" "hand-cursor.xbm")
                                  'xbm (make-object color% 255 255 255))
                                6 2)
                              (make-object cursor% 
                                (make-object bitmap% 
                                  (build-path "images" "hand-cursor.xbm")
                                  'xbm (make-object color% 255 255 255))
                                (make-object bitmap% 
                                  (build-path "images" "hand-cursor-mask.xbm")
                                  'xbm (make-object color% 255 255 255))
                                6 2)))
    ;; Here is the bitmap and the associated dc used in the on-paint method.
    (define bitmap (make-object bitmap% 510 259 #f))
    (define dc (make-object bitmap-dc% bitmap))
    
    ;; These variable save the status of each link...
    (define e-mail-address-focused #f)
    (define website-address-focused #f)
    
    ;; We define the main picture...
    (define k-logo  (make-object bitmap% (build-path "images" "about.png") 'png #f))
    
    ;; The overriden methods list...
    (override on-event on-paint)

    ;; The on-event method...
    (define on-event (lambda (event) 
                       ;; We assign the coordinates to x and y.
                       (let ((x    (send event get-x))
                             (y    (send event get-y)))
                         ;; What is the event type?
                         (case (send event get-event-type)
                           ;; Is it a motion event?
                           ((motion)
                            ;; Therefore, update the status of the links...
                            (cond ((and (<= 248 x 367) (<= 223 y 234) (not e-mail-address-focused))
                                   (set! e-mail-address-focused #t))
                                  ((and (<= 273 x 436) (<= 236 y 247) (not website-address-focused))
                                   (set! website-address-focused #t))
                                  ((and (not (and (<= 248 x 367) (<= 223 y 234))) e-mail-address-focused)
                                   (set! e-mail-address-focused #f))
                                  ((and (not (and (<= 273 x 436) (<= 236 y 247))) website-address-focused)
                                   (set! website-address-focused #f)))
                            ;; ...update the cursor...
                            (if (or (and (<= 248 x 367) (<= 223 y 234))
                                    (and (<= 273 x 436) (<= 236 y 247)))
                                (send this set-cursor k-hand-cursor)
                                (send this set-cursor k-arrow-cursor))
                            ;; ...and call on-paint.
                            (send this on-paint))
                           
                           ;; Is it a left-down event?
                           ((left-down)
                            ;; Then locate it and call the appropriate URL...
                            (if (and (<= 248 x 367) (<= 223 y 234))
                                (send-url "mailto:jean-pierre@lozi.org"))
                            (if (and (<= 273 x 436) (<= 236 y 247))
                                (send-url "http://mreddesigner.lozi.org")))))))
    
    ;; The on-paint method...
    (define on-paint (lambda ()
                       ;; We draw the logo...
                       (send dc draw-bitmap k-logo 0 0 'solid (make-object color% 0 0 0) #f)
                       ;; ...update the brush/pen...
                       (send dc set-pen k-invisible-pen)
                       (send dc set-brush k-underline-brush)
                       ;; ...underline the focused link (if any)...
                       (if e-mail-address-focused
                           (send dc draw-rectangle 250 233 115 1))
                       (if website-address-focused
                           (send dc draw-rectangle 275 246 159 1))
                       ;; ...and then draw the updated bitmap in the real dc.
                       (send (send this get-dc) draw-bitmap bitmap 0 0 'solid k-black #f)))

    ;; And then we call the default canvas initialization method.
    (super-instantiate ())))
