;; ##################################################################################
;; # ============================================================================== #
;; # MrEd Designer - hierarchy-frame.ss                                             #
;; # http://mreddesigner.lozi.org                                                   #
;; # Copyright (C) Lozi Jean-Pierre, 2004 - mailto:jean-pierre@lozi.org             #
;; # ============================================================================== #
;; #                                                                                #
;; # This program is free software; you can redistribute it and/or                  #
;; # modify it under the terms of the GNU General Public License                    #
;; # as published by the Free Software Foundation; either version 2                 #
;; # of the License, or (at your option) any later version.                         #
;; #                                                                                #
;; # This program is distributed in the hope that it will be useful,                #
;; # but WITHOUT ANY WARRANTY; without even the implied warranty of                 #
;; # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                  #
;; # GNU General Public License for more details.                                   #
;; #                                                                                #
;; # You should have received a copy of the GNU General Public License              #
;; # along with this program; if not, write to the Free Software                    #
;; # Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.    #
;; #                                                                                #
;; ##################################################################################


;; We need to derive the frame class to override the on-close method.
(define hierarchy-frame%
  (class frame%
    ;; We override on-close...
    ;(override on-close)
    ;; When on-close is called, we exit the project.
    (define/augment (on-close) (send properties-frame-windows-hierarchy-frame-menu-item check #f))
    (super-instantiate ())))

;; We instantiate the frame...
(define hierarchy-frame (new hierarchy-frame%
                             (label "Hierarchy")
                             (parent toolbox-frame)
                             (x 5)
                             (y 245)
                             (width (send toolbox-frame get-width))))

;; ...the main horizontal pane...
(define hierarchy-horizontal-panel (new horizontal-panel%
                                        (border 0)
                                        (parent hierarchy-frame)))

;; ...the buttons' panel...
(define buttons-panel (new vertical-panel%
                           (border 1)
                           (alignment '(center top))
                           (stretchable-width #f)
                           (parent hierarchy-horizontal-panel)))

;; ...the swap up button...
(define hierarchy-up-button 
  (new button% 
       (label (make-object 
                  bitmap% 
                (build-path "images" "hierarchy-up.png")
                'png (make-object color% 255 255 255)))
       (callback (lambda (button control-event)
                   (let*
                     ;; get the selected element ...
                     ((selected (get-selected-element project-data-global)))
                     ;; ... swap up the selected element in the project data ...
                     (set! project-data-global 
                           (swap-up-selected project-data-global selected))
                     ;; ... notify the the hierarchy widget ...
                     (send hierarchy-widget set-data project-data-global)
                     ;; ... update the preview when not frames up swapped,
                     ;; when frames are swapped, do nothing ...
                     (let ((parent (project-data-get-parent selected)))
                       (if parent
                         (preview-update-children-order parent)))
                     ;; ... there is nothing to do for the property window
                     ;; as the element is still selected ...
                     ;; ... update the buttons for hierarchy ...
                     (update-hierarchy-buttons selected)
                     ;; ... finally remember the change
                     (set! no-change-since-last-save #f))))
       (enabled #f)
       (parent buttons-panel)))

;; ...the swap up one button...
(define hierarchy-up-one-button 
  (new button% 
       (label (make-object 
                  bitmap% 
                (build-path "images" "hierarchy-up-one.png")
                'png (make-object color% 255 255 255)))
       (callback (lambda (button control-event)
                   (let*
                     ;; get the selected element ...
                     ((selected (get-selected-element project-data-global)))
                     ;; ... swap-up one into ...
                     (set! project-data-global 
                           (swap-up-one-selected project-data-global))
                     ;; ... update preview, it will also change the project data ...
                     ; FIXME: actually if no swap-in occured, for example 
                     ; FIXME: widgets only changed there order in the list then
                     ; FIXME: a less intrusive preview is enough
                     (preview-update-preview project-data-global
                                             project-data-global)
                     ;; ... notify the the hierarchy widget ...
                     (send hierarchy-widget set-data project-data-global)
                     ;; ... update the buttons for hierarchy ...
                     (update-hierarchy-buttons selected)
                     ;; ... finally remember the change
                     (set! no-change-since-last-save #f))))
       (enabled #f)
       (parent buttons-panel)))

;; ...the swap down one button...
(define hierarchy-down-one-button 
  (new button% 
       (label (make-object 
                  bitmap% 
                (build-path "images" "hierarchy-down-one.png")
                'png (make-object color% 255 255 255)))
       (callback (lambda (button control-event)
                   (let*
                     ;; get the selected element ...
                     ((selected (get-selected-element project-data-global)))
                     ;; ... swap-up one into ...
                     (set! project-data-global 
                           (swap-down-one-selected project-data-global))
                     ;; ... update preview, it will also change the project data ...
                     ; FIXME: actually if no swap-in occured, for example 
                     ; FIXME: widgets only changed there order in the list then
                     ; FIXME: a less intrusive preview is enough
                     (preview-update-preview project-data-global
                                             project-data-global)
                     ;; ... notify the the hierarchy widget ...
                     (send hierarchy-widget set-data project-data-global)
                     ;; ... update the buttons for hierarchy ...
                     (update-hierarchy-buttons selected)
                     ;; ... finally remember the change
                     (set! no-change-since-last-save #f))))
       (enabled #f)
       (parent buttons-panel)))


;; ...the swap down button...
(define hierarchy-down-button 
  (new button% 
       (label (make-object 
                  bitmap% 
                (build-path "images" "hierarchy-down.png")
                'png (make-object color% 255 255 255)))
       (callback (lambda (button control-event)
                   (let*
                     ;; get the selected element ...
                     ((selected (get-selected-element project-data-global)))
                     ;; ... swap down the selected element in the project data ...
                     (set! project-data-global 
                           (swap-down-selected project-data-global selected))
                     ;; ... notify the the hierarchy widget ...
                     (send hierarchy-widget set-data project-data-global)
                     ;; ... update the preview,
                     ;; this will only happen if not frames up swapped
                     ;; when frames are swapped, do nothing ...
                     (let ((parent (project-data-get-parent selected)))
                       (if parent
                         (preview-update-children-order parent)))
                     ;; ... there is nothing to do for the property window
                     ;; as the element is still selected ...
                     ;; ... update the buttons for hierarchy ...
                     (update-hierarchy-buttons selected)
                     ;; ... finally remember the change
                     (set! no-change-since-last-save #f))))
       (enabled #f)
       (parent buttons-panel)))

;; ...the delete button...
(define hierarchy-delete-button 
  (new button% 
       (label (make-object 
                  bitmap% 
                (build-path "images" "hierarchy-delete.png")
                'png (make-object color% 255 255 255)))
       (callback (lambda (button control-event)
                   (let
                     ;; get the selected element ...
                     ((selected (get-selected-element project-data-global)))
                     ;; ... delete it from the project data ...
                     (set! project-data-global
                           (delete-selected project-data-global selected))
                     ;; ... notify the the hierarchy widget ...
                     (send hierarchy-widget set-data project-data-global)
                     ;; ... then, delete the element in the preview ...
                     (preview-delete-element selected)
                     ;; ... empty out the property window ...
                     (send properties-widget empty)
                     ;; ... update the buttons for hierarchy ...
                     (update-hierarchy-buttons #f)
                     ;; ... finally remember the change
                     (set! no-change-since-last-save #f))))
       (enabled #f)
       (parent buttons-panel)))
  
;; ...the hide button...
(define hierarchy-hide-button 
  (new button% 
       (label (make-object 
                  bitmap% 
                (build-path "images" "hierarchy-hide.png")
                'png (make-object color% 255 255 255)))
       (callback (lambda (button control-event)
                   (let ((selected (get-selected-element project-data-global)))
                     (send (get-property-value selected 'pointer) show #f)
                     (update-hierarchy-buttons selected))))
       (enabled #f)
       (parent buttons-panel)))

;; ...the show button...
(define hierarchy-show-button 
  (new button% 
       (label (make-object 
                  bitmap% 
                (build-path "images" "hierarchy-show.png")
                'png (make-object color% 255 255 255)))
       (callback (lambda (button control-event)
                   (let ((selected (get-selected-element project-data-global)))
                     (send (get-property-value selected 'pointer) show #t)
                     (update-hierarchy-buttons selected))))
       (enabled #f)
       (parent buttons-panel)))

;; the hierarchy widget itself.
(define hierarchy-widget 
  (new hierarchy-widget% 
       (parent hierarchy-horizontal-panel)
       (callback 
        (lambda (action data selected)
          ; store the data
          (set! project-data-global data)
          ; if something was selected
          (if (eq? action 'select)
            (begin
              ;; Update the property widget
              (send properties-widget update-in)
              ;; Update the buttons in the properties widget
              (send properties-widget update-buttons)))
          ;; Update hierarchy widgets
          (update-hierarchy-buttons selected)))))

;; This function updates the enabled/disabled buttons status.
(define (update-hierarchy-buttons selected-element)
  (if selected-element
    (let ((id (get-property-value selected-element 'id)))
      ; if a frame (window) is selected
      (if (eq? (car selected-element) 'frame)
        ;; Ok this is a frame. Is it shown?
        (if (send (get-property-value selected-element 'pointer) is-shown?)
          (begin
            (send hierarchy-show-button enable #f)
            (send hierarchy-hide-button enable #t))
          (begin 
            (send hierarchy-show-button enable #t)
            (send hierarchy-hide-button enable #f)))
        ;; This is not a frame. Therefore disable both buttons.
        (begin
          (send hierarchy-show-button enable #f)
          (send hierarchy-hide-button enable #f)))
      ; if the project is selected
      (if (eq? (car selected-element) 'project)
        (send hierarchy-delete-button enable #f)
        (send hierarchy-delete-button enable #t))
      ; check swapping states
      (if (can-swap-down-selected? project-data-global id)
        (send hierarchy-down-button enable #t)
        (send hierarchy-down-button enable #f))
      (if (can-swap-down-one-selected? project-data-global)
        (send hierarchy-down-one-button enable #t)
        (send hierarchy-down-one-button enable #f))
      (if (can-swap-up-one-selected? project-data-global)
        (send hierarchy-up-one-button enable #t)
        (send hierarchy-up-one-button enable #f))
      (if (can-swap-up-selected? project-data-global id)
        (send hierarchy-up-button enable #t)
        (send hierarchy-up-button enable #f)))
    (begin
      (send hierarchy-up-button enable #f)
      (send hierarchy-up-one-button enable #f)
      (send hierarchy-down-one-button enable #f)
      (send hierarchy-down-button enable #f)
      (send hierarchy-delete-button enable #f)
      (send hierarchy-show-button enable #f)
      (send hierarchy-hide-button enable #f))))



;; We want to be sure that the frame's virtual size is valid before calling update-in...
(send hierarchy-frame reflow-container)

;; We update the widget data... 
(send hierarchy-widget set-data project-data-global)
;; update the buttons
(update-hierarchy-buttons (get-selected-element project-data-global))

;; We show the hierarchy frame...
(send hierarchy-frame show #t)

;; FIXME: does not belong here
;; And then we send a show event to the toolbox frame so that it is focused on startup...
; (send toolbox-frame show #t)
