;; ##################################################################################
;; # ============================================================================== #
;; # MrEd Designer - hierarchy-widget.ss                                            #
;; # http://mreddesigner.lozi.org                                                   #
;; # Copyright (C) Lozi Jean-Pierre, 2004 - mailto:jean-pierre@lozi.org             #
;; # ============================================================================== #
;; #                                                                                #
;; # This program is free software; you can redistribute it and/or                  #
;; # modify it under the terms of the GNU General Public License                    #
;; # as published by the Free Software Foundation; either version 2                 #
;; # of the License, or (at your option) any later version.                         #
;; #                                                                                #
;; # This program is distributed in the hope that it will be useful,                #
;; # but WITHOUT ANY WARRANTY; without even the implied warranty of                 #
;; # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                  #
;; # GNU General Public License for more details.                                   #
;; #                                                                                #
;; # You should have received a copy of the GNU General Public License              #
;; # along with this program; if not, write to the Free Software                    #
;; # Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.    #
;; #                                                                                #
;; ##################################################################################

(module mreddesigner-hierarchy-operation mzscheme

  (require (file "mreddesigner-function.ss")
           (file "mreddesigner-hierarchy-widget.ss")   ; for container?
           (lib "mred.ss" "mred")         ; for warning message
  )
  
  (provide get-selected-element
           get-selected-element-id
           ;;;
           can-swap-up-selected?
           can-swap-down-selected?
           can-swap-up-one-selected?
           can-swap-down-one-selected?
           ;;;
           swap-up-selected
           swap-down-selected
           swap-up-one-selected
           swap-down-one-selected
           ;;;
           delete-selected
  )
  
  ;; FIXME: it should search only until it finds the selected element !!!
  ;; Returns the selected element
  (define (get-selected-element data)
    (letrec
      ((selected #f)
       (select-func (lambda (line)
                      (if (and (get-property-value line 'selected)
                               (not (eq? (car line) 'project)))
                        (set! selected line)))))
      (apply-procedure-on-data-list data select-func)
      selected
    )
  )
    
  ;; Returns the selected element id
  (define (get-selected-element-id data)
    (cadr (assq 'id (cdr (get-selected-element data)))))

  ; removes the child with id from the list
  (define (remove-element-with-id data-list id)
    (cond ( (null? data-list) 
            data-list)
          ( (eq? (get-property-value (car data-list) 'id) id)
            (cdr data-list)
          )
          ( else 
            (cons (car data-list) (remove-element-with-id (cdr data-list) id)))))
    
  ; Inserts an object before an object
  (define (insert-element-before data id obj)
    (cond ((null? data) data)
          ((eq? (get-property-value (car data) 'id) id)
           ;; Then insert before
           (append (list obj (car data)) (cdr data)))
           ;; Otherwise, search in the cdr.
          (else (cons (car data) (insert-element-before (cdr data) id obj)))))

  ; Inserts an object after an object
  (define (insert-element-after data id obj)
    (cond ((null? data) data)
          ((eq? (get-property-value (car data) 'id) id)
           ;; Then insert after
           (append (list (car data) obj) (cdr data)))
           ;; Otherwise, search in the cdr.
          (else (cons (car data) (insert-element-after (cdr data) id obj)))))
  
  ; --------------------------------------------------------------------------
  ;                               Simply swap elements
  ; --------------------------------------------------------------------------
  
  ;; This function returns true iff the element given by 'id' is in the data list.
  (define (in-list? data id)
    ;; If the list is empty, it does not contain the element.
    (cond ((null? data) #f)
          ;; If the first element is the element given by 'id', return true...
          ((eq? (get-property-value (car data) 'id) id) #t)
          ;; Otherwise, search in the cdr.
          (else (in-list? (cdr data) id))))
    
  ;; This function returns true iff the element given by 'id' is the first
  ;; in the list.
  (define (first-in-list? data id)
    ;; We just check the first element...
    (eq? (get-property-value (car data) 'id) id))
    
  ;; This function returns #t iff the element given by 'id' is the last one
  ;; in the data list, assuming that the element is in the list.
  (define (last-in-list? data id)
    ;; If we found the element, just look whether it is the last one...
    (cond ((eq? (get-property-value (car data) 'id) id) (null? (cdr data)))
          ;; Otherwise, search in the cdr.
          (else (last-in-list? (cdr data) id))))
    
  ;; This function swaps the element given by 'id' with the previous one,
  ;; assuming that the element is in the list and is not the first one.
  (define (swap-element-with-previous data id)
    ;; Did we find the element given by 'id'?
    (cond ((eq? (get-property-value (cadr data) 'id) id)
           ;; Then swap it with the previous one!
           (append (list (cadr data)) (list (car data)) (cddr data)))
           ;; Otherwise, search in the cdr.
          (else (cons (car data) (swap-element-with-previous (cdr data) id)))))
    
  ;; This function swaps the element given by 'id' with the next one,
  ;; assuming that the element is in the list and is not the last one.
  (define (swap-element-with-next data id)
    ;; Did we find the element given by 'id'?
    (cond ((eq? (get-property-value (car data) 'id) id)
           ;; Then swap it with the next one!
           (append (list (cadr data)) (list (car data)) (cddr data))
          )
          ;; Otherwise, search in the cdr.
          (else (cons (car data) (swap-element-with-next (cdr data) id)))))
        
  ;; This function returns #t iff the currently selected element can be swapped (up or down,
  ;; depending on the condition).
  (define (can-swap-selected? data id cannot-swap-condition?)
    ;; This helper procedure will be applied to each line...
    (letrec 
      ((can-swap-selected #f)
       (check-func (lambda (line)
                     ;; First we get the children list...
                     (let ((children-list (get-property-children line)))
                       ;; If we found one...
                       ;;  We search whether the element given by 'id' is in the list...
                       (if (and children-list (in-list? children-list id))
                         ;; And then update the result...
                         (set! can-swap-selected 
                               (not (cannot-swap-condition? children-list id))))))))
      ;; We apply the procedure helper on each element of the data list...
      (apply-procedure-on-data-list data check-func)
      ;; And return the result...
      can-swap-selected))
    
  ;; This function returns #t iff the currently selected element can be swapped down.
  (define (can-swap-down-selected? data id)
    ;; The element can be swapped iff it is not the last one in its brother's list.
    (can-swap-selected? data id last-in-list?))
    
  ;; This function returns #t iff the currently selected element can be swapped up.
  (define (can-swap-up-selected? data id)
    ;; The element can be swapped iff it is not the first one in its brother's list.
    (can-swap-selected? data id first-in-list?)) 
    
  ;; This function swaps (up or down depending on the swap function) 
  ;; the currently selected element...
  (define (swap-selected data id swap-function)
    ;; This helper procedure will be applied to each line...
    (letrec ((helper (lambda (line)
                     ;; First we get the children list...
                     (let ((children-list (get-property-children line)))
                       ;; If we found one...
                       (if children-list
                           ;; If the currently selected element is in the list...
                           (if (in-list? children-list id)
                               ;; Then swap it with the given function! 
                               ;; And return the updated line...
                               (set-property-children line (swap-function children-list id))
                               ;; Otherwise do not change the line...
                               line)
                           ;; Otherwise do not change the line...
                           line)))))
      ;; return the data
      (apply-function-on-data-list data helper)
    )
  )
    
  ;; This function swaps up the currently selected element...
  (define (swap-up-selected data selected)
    ;; We call swap-selected with the swap-element-with-previous function.
    (swap-selected data (get-property-value selected 'id) 
                   swap-element-with-previous)
  )
      
  ;; This function swaps down the currently selected element...
  (define (swap-down-selected data selected)
    ;; We call swap-selected with the swap-element-with-next function.
    (swap-selected data (get-property-value selected 'id) 
                   swap-element-with-next)
  )
    
  ; --------------------------------------------------------------------------
  ;                               Swap elements into
  ; --------------------------------------------------------------------------
  
  (define (can-move-above? obj pparent parent prev first? data-list)
    (cond
      ((null? data-list) #f)
      ((get-property-value (car data-list) 'selected)
       ; the condition is here
               ; cannot swap high level into another one
       (if (or (eq? (caar data-list) 'frame)
               (eq? (caar data-list) 'project)
               ; cannot swap first thing in the project
               (and first? parent      (eq? (car parent) 'project))
               ; cannot swap out of a high level into the project
               (and (not prev) pparent (eq? (car pparent) 'project))
           )
         #f #t)
      )
      (else
       (let ((res #f))
         (if (assq 'children (cdar data-list))
           (set! res (can-move-above? #f ; obj
                                      parent ; pparent
                                      (car data-list) ; parent
                                      prev ;prev
                                      #t   ; first?
                                      (cdr (assq 'children (cdar data-list)))))
         )
         (if (not res)
             (can-move-above? #f ; obj
                              pparent ; pparent
                              parent  ; parent
                              (car data-list) ;prev
                              #f  ; first?
                              (cdr data-list))
             res
         )
       )
      )
    )
  )
    
  (define (can-move-below? obj pparent parent next last? data-list)
    (cond
      ((null? data-list) #f)
      ((get-property-value (car data-list) 'selected)

       (let ((last? (null? (cdr data-list))))
         (if (not last?)
           (set! next (cadr data-list))
         )
         ; the condition is here
                 ; cannot swap high level into another one
         (if (or (eq? (caar data-list) 'frame)
                 (eq? (caar data-list) 'project)
                 ; cannot swap last thing in the project
                 (and last? parent      (eq? (car parent) 'project))
                 ; cannot swap out of a high level into the project
                 (and (not next) pparent (eq? (car pparent) 'project))
             )
           #f #t)
       )
      )
      (else
       (let* 
         ((res #f)
          (children-list (assq 'children (cdar data-list)))
         )
         (if children-list
           (set! res (can-move-below? #f ; obj
                                      parent ; pparent
                                      (car data-list) ; parent
                                      ;next
                                      (if (null? (cdr data-list))
                                        next
                                        (cadr data-list)
                                      )
                                      #f
                                      (cdr children-list)))
         )
         (if (not res)
           (can-move-below? #f ; obj
                            pparent ; pparent
                            parent  ; parent
                            next ;next
                            #f
                            (cdr data-list))
           res
         )
       )
      )
    )
  )
  
  ;; This function returns #t iff the currently selected element can be swapped up one.
  (define (can-swap-up-one-selected? data)
    (can-move-above? #f #f #f #f #f data)
  )

  ;; This function returns #t iff the currently selected element can be swapped down one.
  (define (can-swap-down-one-selected? data)
    (can-move-below? #f #f #f #f #f data)
  )
    
  ; It collects all data for moving a widget upwards in the hierarchy:
  ; obj - the widget data
  ; pparent - parent of the parent of the current widget
  ; parent - parent of the current widget
  ; prev - if the widget is the first in its parent then
  ;        it is the previous widget before the parent
  ;        otherwise it is the previous widget in the parent
  ; first? - whether it was the first widget in the parent
  (define (get-data-for-up obj pparent parent prev first? data-list)
    (cond
      ((null? data-list) #f)
      ((get-property-value (car data-list) 'selected)
       (list (car data-list) pparent parent prev first?))
      (else
       (let ((res #f))
         (if (assq 'children (cdar data-list))
           (set! res (get-data-for-up #f ; obj
                                      parent ; pparent
                                      (car data-list) ; parent
                                      prev ;prev
                                      #t   ; first?
                                      (cdr (assq 'children (cdar data-list))) ))
         )
         (if res res (get-data-for-up #f ; obj
                                      pparent ; pparent
                                      parent  ; parent
                                      (car data-list) ;prev
                                      #f  ; first?
                                      (cdr data-list) ))
       )
      )
    )
  )
  
  ; It collects all data for moving a widget downwards in the hierarchy:
  ; obj - the widget data
  ; pparent - parent of the parent of the current widget
  ; parent - parent of the current widget
  ; next - if the widget is the last in its parent then
  ;        it is the next widget after the parent
  ;        otherwise it is the next widget in the parent
  ; last? - whether it was the last widget in the parent
  (define (get-data-for-down obj pparent parent next last? data-list)
    (cond
      ((null? data-list) #f) ;if null, then we have not found it
      ; if id matches return data
      ((get-property-value (car data-list) 'selected)
       (let ((last? (null? (cdr data-list))))
         (list (car data-list) pparent parent (if last? next (cadr data-list)) last?)
       )
      )
      (else
       (let ((res #f))
         (if (assq 'children (cdar data-list))
           (set! res (get-data-for-down #f ; obj
                                        parent ; pparent
                                        (car data-list) ; parent
                                        ;next
                                        (if (null? (cdr data-list))
                                          next
                                          (cadr data-list)
                                        )
                                        #f
                                        (cdr (assq 'children (cdar data-list))) ))
         )
         (if res res (get-data-for-down #f ; obj
                                        pparent ; pparent
                                        parent  ; parent
                                        next ;next
                                        #f
                                        (cdr data-list) ))
       )
      )
    )
  )
  
  ; moves one item up in the hierarchy
  (define (move-up-func data-list)
    ; collect all data
    (let*
      ((info    (get-data-for-up #f #f #f #f #f data-list))
       (obj     (list-ref info 0))
       (pparent (list-ref info 1))
       (parent  (list-ref info 2))
       (prev    (list-ref info 3))
       (first?  (list-ref info 4))
       (id      (get-property-value obj 'id))
       (func-delete
         (lambda (line)
           (let ((children-list (get-property-children line)))
             (if (and children-list (in-list? children-list id))
               (set-property-children line (remove-element-with-id children-list id))
               line
             )
           )
         )
       )
       (func-insert-1
         (lambda (line)
           (if (eq? (get-property-value line 'id) (get-property-value prev 'id))
             ; prev is a container, so simply add to it
             (let ((children-list (get-property-children line)))
               (set! obj (set-property-parent obj (get-property-value prev 'id)))
               ; it is possible that there is no children
               (if children-list
                 (set-property-children line (append children-list (list obj)))
                 (append (list (car line)) (list (list 'children obj)) (cdr line))
               )
             )
             line)))
       (func-insert-2
         (lambda (line)
           ; first, but prev is not a container then add to pparent before parent
           (if (eq? (get-property-value line 'id) (get-property-value pparent 'id))
             (begin
               (set! obj (set-property-parent obj (get-property-value pparent 'id)))
               (set-property-children line (insert-element-before (get-property-children line) 
                                                                  (get-property-value parent 'id) 
                                                                  obj))
             )
             line)))
       (func-insert-3
         (lambda (line)
           ; not first, there is prev but not a container then add before prev
           (if (eq? (get-property-value line 'id) (get-property-value parent 'id))
             (set-property-children line (insert-element-before (get-property-children line) 
                                                                (get-property-value prev 'id) 
                                                                obj))
             line)))
      )
        
;        ; for debugging
;        (display "\nobj: ")(display obj)
;        (display "\n\npparent: ")(display pparent)
;        (display "\n\nparent: ")(display parent)
;        (display "\n\nprev: ")(display prev)
;        (display "\n\nfirst?: ")(display first?)
        
      ;delete from current position
      (set! data-list (apply-function-on-data-list data-list func-delete))
      
      (cond
        ; Current was the first in the list, but 
        ; we cannot swap into the project a non-high-level object.
        ; If everything is OK add current to parent of the parent of the current, 
        ; just before the parent.
        ( (and first? pparent (not (eq? (car pparent) 'project)))
          (apply-function-on-data-list data-list func-insert-2)
        )
        ; the previous widget is a container so add the current to it
        ( (and prev
               (container? prev)
          )
          (apply-function-on-data-list data-list func-insert-1)
        )
        ; current was not the first in the list, therefore there is a prev
        ; and add current before prev
        ( else
          (apply-function-on-data-list data-list func-insert-3)
        )
      )
    )
  )

  ; moves one item down in the hierarchy
  (define (move-down-func data-list)
    ; collect all data
    (let*
      ((info    (get-data-for-down #f #f #f #f #f data-list))
       (obj     (list-ref info 0))
       (pparent (list-ref info 1))
       (parent  (list-ref info 2))
       (next    (list-ref info 3))
       (last?   (list-ref info 4))
       (id      (get-property-value obj 'id))
       (func-delete
         (lambda (line)
           (let ((children-list (get-property-children line)))
             (if (and children-list (in-list? children-list id))
               (set-property-children line (remove-element-with-id children-list id))
               line
             )
           )
         )
       )
       (func-insert-1
         (lambda (line)
           (if (eq? (get-property-value line 'id) (get-property-value next 'id))
             ; next is a container, so simply add to it
             (let ((children-list (get-property-children line)))
               (set! obj (set-property-parent obj (get-property-value next 'id)))
               ; it is possible that there is no children
               (if children-list
                 (set-property-children line (cons obj children-list))
                 (append (list (car line)) (list (list 'children obj)) (cdr line))
               )
             )
             line)))
       (func-insert-2
         (lambda (line)
           ; last, but next is not a container then add to pparent after parent
           (if (eq? (get-property-value line 'id) (get-property-value pparent 'id))
             (begin
               (set! obj (set-property-parent obj (get-property-value pparent 'id)))
               (set-property-children line (insert-element-after (get-property-children line)
                                                                 (get-property-value parent 'id)
                                                                 obj))
             )
             line)))
       (func-insert-3
         (lambda (line)
           ; not last, there is next but not a container then add after next
           (if (eq? (get-property-value line 'id) (get-property-value parent 'id))
             (set-property-children line (insert-element-after (get-property-children line)
                                                               (get-property-value next 'id) 
                                                               obj))
             line)))
      )
      
;        ; for debugging
;        (display "\nobj: ")(display obj)
;        (display "\npparent: ")(display pparent)
;        (display "\nparent: ")(display parent)
;        (display "\nnext: ")(display next)
;        (display "\nlast?: ")(display last?)
        
      ;delete from current position
      (set! data-list (apply-function-on-data-list data-list func-delete))
      
      (cond
        ; current was the last in the list, therefore add current to the
        ; parent of the parent of the current, just after the parent
        ( (and last? pparent (not (eq? (car pparent) 'project)))
          (apply-function-on-data-list data-list func-insert-2)
        )
        ; the next widget is a container so add the current to it
        ( (and next
               (container? next)
          )
          (apply-function-on-data-list data-list func-insert-1)
        )
        ; current was not the last in the list, therefore there is a next
        ; and add current after next
        ( else
          (apply-function-on-data-list data-list func-insert-3)
        )
      )        
    )
  )
        
  ;; This function swaps up the currently selected element...
  (define (swap-up-one-selected data)
    (move-up-func data)
  )
      
  ;; This function swaps down the currently selected element...
  (define (swap-down-one-selected data)
    (move-down-func data)
  )
  
  ; --------------------------------------------------------------------------
  ;                               Delete element
  ; --------------------------------------------------------------------------
  
  ;; Returns number of first-level children
  (define (children-num data-list)
    (let ((children-list (assq 'children (cdr data-list))))
      (if children-list
        (length (cdr children-list))
        0)))
  
  ;; Deletes the selected element
  (define (delete-selected data selected)
    (let* ((id (get-property-value selected 'id))
           (ok? #t)
           (del-line? (lambda (line)
                        (if (and (assq 'id (cdr line))
                            ;; Is the element selected??
                            (eq? (cadr (assq 'id (cdr line))) id))
                          'nil
                          line)))
          )
      ;; We provide a different feedback depending whether the element which is about
      ;; to be deleted has children...
      (if (> (children-num selected) 0)
        (if (not (eq? 'ok 
                      (message-box "Warning" 
                                   (string-append 
                                     "This element has children.\n" 
                                     "Deleting it will cause the loss of all its children.\n"
                                     "Are you sure you want to continue?")
                                   #f
                                   '(ok-cancel caution))))
          (set! ok? #f)))
      
      (if ok?
        (begin
          ;; delete the element
          (apply-function-on-data-list-improved data del-line?)
        )
        #f
      )
    )
  )
  
  
) ; end of module


