;; ##################################################################################
;; # ============================================================================== #
;; # MrEd Designer - hierarchy-widget.ss                                            #
;; # http://mreddesigner.lozi.org                                                   #
;; # Copyright (C) Lozi Jean-Pierre, 2004 - mailto:jean-pierre@lozi.org             #
;; # ============================================================================== #
;; #                                                                                #
;; # This program is free software; you can redistribute it and/or                  #
;; # modify it under the terms of the GNU General Public License                    #
;; # as published by the Free Software Foundation; either version 2                 #
;; # of the License, or (at your option) any later version.                         #
;; #                                                                                #
;; # This program is distributed in the hope that it will be useful,                #
;; # but WITHOUT ANY WARRANTY; without even the implied warranty of                 #
;; # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                  #
;; # GNU General Public License for more details.                                   #
;; #                                                                                #
;; # You should have received a copy of the GNU General Public License              #
;; # along with this program; if not, write to the Free Software                    #
;; # Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.    #
;; #                                                                                #
;; ##################################################################################

(module mreddesigner-hierarchy-widget mzscheme

  (require (lib "class.ss")  
           (lib "mred.ss" "mred")
           (file "mreddesigner-function.ss")
  )

  (provide container?
           hierarchy-widget%
  )

  ;; This predicate returns #t if the line is a container
  (define (container? line)
    (and (pair? line)
         (or (eq? (car line) 'frame)
             (eq? (car line) 'pane)
             (eq? (car line) 'panel)
             (eq? (car line) 'horizontal-pane)
             (eq? (car line) 'horizontal-panel)
             (eq? (car line) 'vertical-pane)
             (eq? (car line) 'vertical-panel)
             (eq? (car line) 'tab-panel)
             (eq? (car line) 'group-box-panel)
             (eq? (car line) 'grow-box-spacer-pane))))

; ------------------------------------------------------------------------------
; ------------------------------------------------------------------------------
;                                Hierarchy widget
; ------------------------------------------------------------------------------
; ------------------------------------------------------------------------------


;;
;; This is the hierarchy-widget file.
;; We derive a canvas, and then code the whole widget from scratch :)
;;
 
;; We declare the class...
(define hierarchy-widget%
  ;; We derive it...
  (class canvas%

    ;; Overriden methods
    (override on-scroll on-size on-paint on-event on-char)

    ;; Public fields
    (init-field (callback #f))
    ; check the callback function
    (unless (or (not callback) 
                (procedure-arity-includes? callback 3))
      (raise-type-error 'hierarchy-widget%
                        "procedure of arity 3"
                        callback)
    )
    
    ;; Public methods
    (public set-data
            get-data
    )

    ; --------------------------------------------------------------------------
    ;                             Local variables, fields
    ; --------------------------------------------------------------------------

    ;; Fields
    (define data '())
    ;; Why using a bitmap and a drawing context as a class field instead
    ;; of directly declare it in the on-paint body? The matter is,
    ;; that if we did so a bitmap would be instantiated for each call of
    ;; the on-paint method. After testing it, even on a pretty powerful
    ;; pc, if the on-paint method is called too often, the program runs out
    ;; of resources and returns an error when instantiating a bitmap after
    ;; a short while.
    (define bitmap #f)
    (define dc #f)
    (define x-offset 0)
    (define y-offset 0)
    
    ;; Constants
    (define k-left-margin 21)
    (define k-top-margin 2)
    (define k-line-height 17)
    (define k-horizontal-scroll-step 3)
    
    ;; Those constants are used for drawing
    (define k-white         (make-object color% 255 255 255))
    (define k-black         (make-object color% 0 0 0))
    (define k-gray          (make-object color% 160 160 160))
    (define k-white-brush   (make-object brush% k-white 'solid))
    (define k-active-brush  (make-object brush% k-black 'hilite))
    ;; We use the default panel background (which depends on the platform, windowing system, etc.)
    (define k-gray-brush    (make-object brush% (get-panel-background) 'solid))
    ;; Opaque is needed here
    (define k-dkgray-brush  (make-object brush% (make-object color% 132 132 132) 'opaque))
    (define k-invisible-pen (make-object pen%   (make-object color% 0 0 0) 0 'transparent))
    (define k-edit-font     (make-object font%  12 "Courier" 'default 'normal 'normal #f 
                                                'unsmoothed #t))
    
    ;; To be platform compliant, we avoid smoothed text under windows, because highlighted smoothed
    ;; text under windows does not look really nice.
    (define k-normal-font        (if (eq? (system-type) 'windows)
                                   (make-object font% 11 'system 'normal 'normal #f 'unsmoothed #t)
                                   (make-object font% 11 'system 'normal 'normal #f 'smoothed #t)))
    
    (define k-italic-font        (if (eq? (system-type) 'windows)
                                   (make-object font% 11 'system 'italic 'normal #f 'unsmoothed #t)
                                   (make-object font% 11 'system 'italic 'normal #f 'smoothed #t)))
    
    (define k-bold-font          (if (eq? (system-type) 'windows)
                                   (make-object font% 11 'system 'normal 'bold #f 'unsmoothed #t)
                                   (make-object font% 11 'system 'normal 'bold #f 'smoothed #t)))
    
    ;; Loading the bitmaps now will allow us to reduce time needed to draw the widget.
    (define k-less-bitmap (make-object bitmap% (build-path "images" "less.png") 'png #f))
    (define k-plus-bitmap (make-object bitmap% (build-path "images" "plus.png") 'png #f))
    (define k-project-mask-bitmap (make-object bitmap% 
                                               (build-path "images" "icons" "16x16" "project-mask.png")
                                               'png #f))
    ;; Why using a a-list instead of declaring each icon? 
    ;; Of course declaring each icon would be a little faster,
    ;; but by using a a-list we can add as many icons as needed without changing the code.
    (define k-icon-list 
      (map (lambda (elt) (cons (string->symbol (substring elt 0 (- (string-length elt) 4))) 
                               (make-object bitmap% 
                                 (build-path "images" "icons" "16x16" elt)
                                 'png #f)))
           ;; We look for the icons in the right directory...
           (map (lambda (p) (path->string p))
                (directory-list (path->string (build-path "images" "icons" "16x16"))))
      )
    )    
    
    ; --------------------------------------------------------------------------
    ;                                Operations
    ; --------------------------------------------------------------------------
    
    ;; the function returns the internal data representation
    (define (get-data)
      data
    )
    
    ;; This function updates the local environment with the contents of the global environment...
    (define (set-data outside-data)
      ;; ...we first update the data list...
      (set! data outside-data)
      ;; ...the bitmap and scrollbars ...
      (update-bitmap)
      ;; ...and finally repaint the whole widget.
      (on-paint))
    
    
    ;; Quite self-explanatory...
    (define (collapsed? data-list)
      (cadr (assq 'collapsed (cdr data-list))))
        
    ; --------------------------------------------------------------------------
    ;                      Update canvas and scrollbars
    ; --------------------------------------------------------------------------
    
    ;; This function returns the "real" width of the widget, i.e.
    ;; the total width needed to display the whole widget's contents.
    (define (get-real-width)
      ;; This function will parse the whole list in order to find the line
      ;; which needs the more pixel (in width) to be drawn...
      (define (parse-data-list data-list offset)
        ;; The maximum has been saved into previous-max. Therefore we just need to return
        ;; previous-max if the list parsing has ended (that is, the list is empty).
        (if (null? data-list)
            0 
            ;; Otherwise, the element we are working with is the car of the list.
            (let ((current-element (car data-list)))
              ;; Then we are going to simulate the drawing in order to find the line width (in pixels).
              (let-values (((label-width label-height label-distance label-extra)
                            (get-label-extent current-element)
                           )
                           ;; Ok, we found the label width - we need to check the id width, now.
                           ((id-width id-height id-distance id-extra)
                            (get-id-extent current-element)
                           ))
                ;; We are now able to find the length of the current line, and therefore save it...
                (let* ((current-length (+ (* 20 offset) k-left-margin 20 label-width id-width 10)))
                  ;; Do the current element have children?
                  (if (and (assq 'children (cdr current-element))
                           (not (collapsed? current-element)))
                    ;; If it has, consider the maximum line length of children, 
                    ;; in the maximum evaluation...
                    (max (parse-data-list (cdr (assq 'children (cdr current-element))) (add1 offset))
                         (parse-data-list (cdr data-list) offset)
                         current-length)
                    ;; Otherwise, only consider the maximum line length of brothers 
                    ;; (and of course the current line length).
                    (max (parse-data-list (cdr data-list) offset)
                         current-length)))))))
      
      ;; And then, we call the function we just declared...
      (inexact->exact (parse-data-list data 0)))
    
    ;; This function returns the "real" height of the widget, i.e.
    ;; the total height needed to display the whole widget's contents.
    (define (get-real-height)
      ;; We add 1 to ensure the height is != 0...
      (add1 (* k-line-height (get-number-of-lines))))
        
    ;; Returns total number of children
    (define (children-num-r data-list)
      (cond ((null? data-list) 0)
            ;; No children?
            ((not (assq 'children (cdr data-list))) 1)
            (else (add1 (apply + (map (lambda(elt) (children-num-r elt))
                                      (cdr (assq 'children (cdr data-list)))))))))
    
    ;; Returns total number of non-collapsed children
    (define (children-num-r-nc data-list)
      (cond ((null? data-list) 0)
            ;; No children?
            ((not (assq 'children (cdr data-list))) 1)
            ;; Collapsed?
            ((collapsed? data-list) 1)
            ;; otherwise...
            (else (add1 (apply + (map (lambda(elt) (children-num-r-nc elt))
                                      (cdr (assq 'children (cdr data-list)))))))))
    
    ;; Returns the number of lines in the widget
    (define (get-number-of-lines)
      (if (null? data)
          0
          (children-num-r-nc (car data))))
    
    
    ;; This function updates the bitmap in which drawings are performed before
    ;; appearing on the screen.
    (define (update-bitmap)
      (update-scroll-bars)
      ;; We will need the virtual size...
      (let-values (((width height) (send this get-virtual-size)))
        ;; We create a bitmap, with the *real* width and height (at least, if they are
        ;; greater than the virtual width and height...), in order to be able to scroll very quickly...
        (set! bitmap (make-object bitmap% (max (get-real-width) width 1) 
                                          (max (get-real-height) height 1) #f)))
      ;; And then, we update the associated bitmap-dc...
      (set! dc (instantiate bitmap-dc% (bitmap)))
    )

    ;; This function updates the scroll bars properties...
    (define (update-scroll-bars)
      (let
        ((horizontal? #f) (vertical? #f))
        ;; hide both of the scroll bars
        (send this show-scrollbars #f #f)
        ;; We need to get the virtual size...
        (let-values (((width height) (send this get-virtual-size)))
          ;; Then, we update the scroll range for both the horizontal and the vertical scroll
          ;; bars. The right range is the full real height (width, respectively) minus the
          ;; virtual height (resp. width).
          ;; vertical scrollbar
          (send this set-scroll-range 'vertical (max 0 (- (get-real-height) height)))
          ;; If the window can contain the whole height, no y offset is needed.
          (if (= 0 (send this get-scroll-range 'vertical)) 
            (set! y-offset 0)
            (begin
              (send this show-scrollbars #f #t)
              (set! vertical? #t)
            )
          )
        )
        (let-values (((width height) (send this get-virtual-size)))
          ;; horizontal scrollbar
          (send this set-scroll-range 'horizontal (max 0 (- (get-real-width) width)))
          ;; If the window can contain the whole width, no x offset is needed.
          (if (= 0 (send this get-scroll-range 'horizontal))
            (set! x-offset 0) 
            (begin
              (send this show-scrollbars #t vertical?)
              (set! horizontal? #t)
            )
          )
        )
        (let-values (((width height) (send this get-virtual-size)))
          ;; check vertical scrolling again
          (send this set-scroll-range 'vertical (max 0 (- (get-real-height) height)))
          ;; If the window can contain the whole height, no y offset is needed.
          (if (= 0 (send this get-scroll-range 'vertical))
            (set! y-offset 0)
            (send this show-scrollbars horizontal? vertical?)
          )
          ;; This is the number of scroll steps - vertically we want to scroll not exactly one
          ;; page, but one page minus one line (as it is always implemented).
          (send this set-scroll-page 'horizontal (max 1 width))
          (send this set-scroll-page 'vertical (max 1 (- height k-line-height)))
        )
      )
    )
        
    ;; +----------------------------------------------------------------------+
    ;; | Function hierarchy-widget::on-scroll                                 |   
    ;; +----------------------------------------------------------------------+
    (define (on-scroll scroll-event)
      ;; What we are going to do depends on the event direction...
      (case (send scroll-event get-direction)
        ;; Is it vertical?
        ((vertical) 
         ;; Then it depends on the event type...
         (let ((type (send scroll-event get-event-type)))
           ;;
           ;; Notice : I am forced to use eq?, because 
           ;; I have noticed a bug under MacOS X : if the event is a thumb event,
           ;; DrScheme crashes if we try to evaluate the result given by get-event-type! 
           ;; It looks to return an invalid pointer, at a system adress, maybe 0. 
           ;; Therefore I just compare pointers.
           ;;
           ;; If the event type is line down...
           (cond ((eq? type 'line-down)
                  ;; ...then we need the virtual size...
                  (let-values (((width height) (send this get-virtual-size)))
                    ;; ...update the scroll-bars positions 
                    ;; (to force scrolling more than 1 step [1 line, actually])...
                    (send this set-scroll-pos 'vertical 
                          (min (+ (send this get-scroll-pos 'vertical) k-line-height -1) 
                               (send this get-scroll-range 'vertical)))
                    ;; ...and the vertical offset.
                    (set! y-offset (min (+ y-offset k-line-height) 
                                        (send this get-scroll-range 'vertical)))))
                 
                 ;; If the event type is line-up...
                 ((eq? type 'line-up)
                  ;; ...then we update the scroll-bars positions 
                  ;; (to force scrolling more than 1 step...)...
                  (send this set-scroll-pos 'vertical 
                        (max (- (send this get-scroll-pos 'vertical) k-line-height -1) 0))
                  ;; ...and the vertical offset.
                  (set! y-offset (max (- y-offset k-line-height) 0)))
                 
                 ;; Otherwise, we do not have to update the scroll-bars positions, 
                 ;; just to update the vertical offset...
                 (else (set! y-offset (send scroll-event get-position))))))
        
        ;; Is it horizontal?
        ((horizontal) 
         ;; Then it depends on the event type...
         (let ((type (send scroll-event get-event-type)))
           ;; If the event type is line down...
           (cond ((eq? type 'line-down)
                  ;; ...then we need the virtual size...
                  (let-values (((width height) (send this get-virtual-size)))
                    ;; ...update the scroll-bars positions 
                    ;; (to force scrolling more than 1 step... [1 "horizontal scroll step", actually])...
                    (send this set-scroll-pos 'horizontal 
                          (min (+ (send this get-scroll-pos 'horizontal) k-horizontal-scroll-step -1) 
                                  (send this get-scroll-range 'horizontal)))
                    ;; ...and the horizontal offset.
                    (set! x-offset (min (+ x-offset k-horizontal-scroll-step) 
                                        (send this get-scroll-range 'horizontal)))))
                 
                 ;; If the event type is line-up...
                 ((eq? type 'line-up)
                  ;; ...then we update the scroll-bars positions 
                  ;; (to force scrolling more than 1 step...)...
                  (send this set-scroll-pos 'horizontal 
                        (max (- (send this get-scroll-pos 'horizontal) k-horizontal-scroll-step -1) 0))
                  ;; ...and the horizontal offset.
                  (set! x-offset (max (- x-offset k-horizontal-scroll-step) 0)))
                 
                 ;; Otherwise, we do not have to update the scroll-bars positions, 
                 ;; just to update the horizontal offset...
                 (else (set! x-offset (send scroll-event get-position)))))))
      
      ;; And then, we directly draw the bitmap, which has already been drawed during the last
      ;; call of (on-paint), at the right position - which is *much* faster than calling on-paint
      ;; directly and allows a smooth scrolling.
      (send (send this get-dc) draw-bitmap bitmap (- x-offset) (- y-offset)))
    
    ;; +----------------------------------------------------------------------+
    ;; | Function hierarchy-widget::on-size                                   |   
    ;; +----------------------------------------------------------------------+
    (define (on-size width height)
      ;; We just update the bitmap...
      (update-bitmap)
    )
    
    ;; +----------------------------------------------------------------------+
    ;; | Function hierarchy-widget::on-paint                                  |
    ;; +----------------------------------------------------------------------+
    
    ;; it returns a string
    (define (get-label-text data-list)
      (if (assq 'label (cdr data-list))
        (let ((label (cadr (assq 'label (cdr data-list)))))
          (if (and label (not (string=? label ""))) label "(no label)"))
        (string-append "<" (symbol->string (car data-list)) ">"))
    )
    
    ;; Determine the extent of the label text
    ;; it returns four values
    (define (get-label-extent data-list)
      (send dc get-text-extent 
            ;; The label text...
            (get-label-text data-list)
            ;; And its font...
            (cond
              ((memq (car data-list) '(frame project)) k-bold-font)
              ((container? data-list) k-italic-font)
              (else k-normal-font)))
    )
    
    ;; it returns a string
    (define (get-id-text data-list)
      (let ((id-v (assq 'id (cdr data-list))))
        (if id-v
          (string-append " [" (symbol->string (cadr (assq 'id (cdr data-list)))) "]") 
          ""
        )
      )
    )
    
    ;; Determine the extent of the ID text
    ;; it returns four values
    (define (get-id-extent data-list)
      (send dc get-text-extent 
            (get-id-text data-list)
            ;; The id is drawn in an italic font...
            k-italic-font)
    )
    
    (define (on-paint)
      (let* ((width  (send this get-width))
             (height (send this get-height)))
        
        ;; Data list parser
        (define (parse-data-list data-list incr offset)
          (if (not (null? data-list))
              (begin
                ;; Text foreground is black...
                (send dc set-text-foreground k-black)
                
                ;; Set the right font...
                (cond ((memq (caar data-list) '(frame project)) (send dc set-font k-bold-font))
                      ((container? (car data-list)) (send dc set-font k-italic-font))
                      (else (send dc set-font k-normal-font)))
                
                ;; The project name is on a gray background.
                (if (zero? offset)
                    (begin
                      (send dc set-brush k-gray-brush)
                      (send dc draw-rectangle 0 
                            (* incr k-line-height) (send bitmap get-width) k-line-height)))
                
                ;; Let's draw the links betweeb the +/- boxes.
                ;; First we choose the color...
                (send dc set-brush k-dkgray-brush)
                
                ;; The lines function parses the children in order to find
                ;; which must and which must not be linked with the current element.
                (letrec ((lines (lambda (n list)
                                  (cond
                                    ((null? list) (void))
                                    (else
                                     (trace n)
                                     (if (collapsed? (car list))
                                       (lines (+ n 1) (cdr list))
                                       (lines (+ n (children-num-r-nc (car list))) (cdr list)))
                                     ))))
                         ;; The trace function actually draws the links.
                         (trace (lambda (n)
                                  (begin
                                    (send dc draw-rectangle 
                                          (+ (* 20 offset) 10) (+ (* (+ n incr) k-line-height) 8)
                                          24 1)
                                    (send dc draw-rectangle 
                                          (+ (* 20 offset) 9)  (+ (* incr k-line-height) 4)
                                          1 (+ (* n k-line-height) 5))))))
                  ;; Then we call the lines function, but only if the current
                  ;; element is not collapsed.
                  (if (not (collapsed? (car data-list)))
                    (let ((children-list (assq 'children (cdar data-list))))
                      (if children-list
                        (lines 1 (cdr children-list))
                        (lines 1 '())))))
                
                (if (assq (caar data-list) k-icon-list)
                  ;; We draw the right icon...
                  (send dc draw-bitmap 
                        ;; We look for the right icon in the a-list...
                        (cdr (assq (caar data-list) k-icon-list)) 
                        (+ (* 20 offset) k-left-margin)
                        (+ (* k-line-height incr) 1) 
                        'solid 
                        (make-object color% "black")
                        (if (eq? (caar data-list) 'project)
                          k-project-mask-bitmap
                          #f)))
                
                ;; Draw the +/- icons
                (if (or (container? (car data-list))
                        (eq? (caar data-list) 'project))
                    ;; If the current element is collapsed...
                    (if (not (collapsed? (car data-list)))
                        ;; ...we draw a +...
                        (send dc draw-bitmap 
                              k-less-bitmap
                              (+ (* 20 offset) 5)
                              (+ (* k-line-height incr) 4) 
                              'solid (make-object color% "white") #f)
                        ;; ...we draw a - otherwise.
                        (send dc draw-bitmap 
                              k-plus-bitmap
                              (+ (* 20 offset) 5)
                              (+ (* k-line-height incr) 4) 
                              'solid (make-object color% "white") #f))) 
                
                (send dc set-text-foreground k-black)
                
                ;; Display the caption...
                (send dc draw-text
                      (get-label-text (car data-list))
                      (+ (* 20 offset) k-left-margin 20) 
                      (+ (* k-line-height incr) k-top-margin))
                
                (let-values (((width height distance extra) (get-label-extent (car data-list))))
                  (send dc set-text-foreground k-gray)
                  
                  ;; Display the id...
                  (send dc set-font k-italic-font)
                  (send dc draw-text
                        (get-id-text (car data-list))
                        (+ (* 20 offset) k-left-margin 20 width)
                        (+ (* k-line-height incr) k-top-margin))
                  
                  ;; Is the current element selected?
                  (if (get-property-value (car data-list) 'selected)
                      ;; Then draw the selection!
                      (begin
                        (send dc set-brush k-active-brush)
                        (send dc draw-rectangle 
                              (+ (* 20 offset) k-left-margin 19) 
                              (+ (* k-line-height incr) distance -1)
                              (+ width 2)
                              (- k-line-height 2)))))
                
                
                ;; Do the current element have children?
                (if (and (assq 'children (cdar data-list))
                         (not (collapsed? (car data-list))))
                    ;; Then show them!
                    (parse-data-list (cdr (assq 'children (cdar data-list))) 
                                     (add1 incr) 
                                     (add1 offset)))
                
                ;; Continue with the rest of the list...
                (parse-data-list (cdr data-list) 
                                 ;; Next line is one plus current line plus number of children
                                 ;; if the element is not collapsed, the current line plus 1
                                 ;; otherwise
                                 (if (and (assq 'children (cdar data-list)) 
                                          (collapsed? (car data-list)))
                                     (+ 1 incr)
                                     (+ 1 incr (sub1 (children-num-r-nc (car data-list)))))
                                 ;; No offset, we are at the same hierarchy level
                                 offset))))
        
        (send dc set-pen k-invisible-pen) ;; Invisible pen
        (send dc set-brush k-white-brush) ;; White brush

        ;; We clear the bitmap...
        (send dc clear)

        ;; ...and then parse the list
        (parse-data-list data 0 0)
        
        ;; We drawed the whole output into a bitmap in order to avoid flickering,
        ;; therefore we have to draw the bitmap.
        (if (send bitmap ok?)
          (send (send this get-dc) draw-bitmap bitmap (- x-offset) (- y-offset))
        )
      )
    )
    
    ;; +----------------------------------------------------------------------+
    ;; | Function hierarchy-widget::on-event                                  |   
    ;; +----------------------------------------------------------------------+
    (define (action-select selected-id)
      (let
        ((select-func (lambda (line)
                        (if (eq? (get-property-value line 'id) selected-id)
                          (set-property-select line #t)
                          (set-property-select line #f)))))
        (set! data (apply-function-on-data-list data select-func))
      )
    )
    
    (define (action-collapse selected-id)
      (let
        ((collapse-func (lambda (line)
                          (if (eq? (get-property-value line 'id) selected-id)
                            (set-property-collapse line #t)
                            line))))
        (set! data (apply-function-on-data-list data collapse-func))
      )
    )

    (define (on-event event)
      ;; The widget gets the focus
      (send this focus)
      
      ;; On-event body starts here
      (let ((x        (+ (send event get-x) x-offset))
            (y        (+ (send event get-y) y-offset))
            (action   #f)
            (selected #f)
           )
        (case (send event get-event-type)
          ;; Mouse-down event with the left mouse button
          ((left-down)
           (letrec 
             ( (parse-data-list 
                 (lambda (data-list incr offset)
                   (if (not (null? data-list))
                     (begin
                       (let-values 
                         (((width height distance extra) (get-label-extent (car data-list))))
                         
                         ;; Is the event a field area?
                         (if (and (>= x (+ (* 20 offset) k-left-margin ))
                                  (<= x (+ (* 20 offset) k-left-margin width 20))
                                  (>= y (+ (* k-line-height incr) distance -1))
                                  (<= y (+ (* k-line-height incr) distance 1 height)))
                           (begin 
                             (set! selected (car data-list))
                             (set! action 'select)))
                                               
                         ;;Is the event in a +/- area?
                         (if (and (>= x (+ (* 20 offset) k-left-margin -16))
                                  (<= x (+ (* 20 offset) k-left-margin -7))
                                  (>= y (+ (* k-line-height incr) 4))
                                  (<= y (+ (* k-line-height incr) 13)))
                           (begin
                             (set! selected (car data-list))
                             (set! action 'collapse)))
                                               
                         (if (and (assq 'children (cdar data-list))
                                  (not (collapsed? (car data-list))))
                           (parse-data-list (cdr (assq 'children (cdar data-list)))
                                            (add1 incr) (add1 offset)))
                         
                         ;; Parse the list
                         (parse-data-list (cdr data-list)
                                          (if (and (assq 'children (cdar data-list))
                                                   (collapsed? (car data-list)))
                                            (+ 1 incr)
                                            (+ 1 incr (sub1 (children-num-r-nc (car data-list))))
                                          )
                                          offset)))))))
             
             ;; Parse the whole list
             (parse-data-list data 0 0)
             (cond
               ((eq? action 'collapse)
                (action-collapse (get-property-value selected 'id))
                (on-paint)
                ;; if an element is uncollapsed then
                ;; the size of the bitmap may grow or vica-versa
                (update-bitmap)
               )
               ((eq? action 'select)
                (action-select (get-property-value selected 'id))
                (on-paint)
               )
             )
             (if (and action callback)
               (callback action data selected)
             )
             
             )))))
    
    ;; +----------------------------------------------------------------------+
    ;; | Function properties-widget::on-char                                  |   
    ;; +----------------------------------------------------------------------+
    (define (on-char event)
      (case (send event get-key-code)
        
        ;; Wheel up
        ((wheel-up)
         (on-scroll (new scroll-event% (event-type 'line-up) (direction 'vertical))))
        
        ;; Wheel down
        ((wheel-down)
         (on-scroll (new scroll-event% (event-type 'line-down) (direction 'vertical))))))
            

    ;; Call the initialization function from the parent class...
    (super-new (style '(vscroll hscroll border)))
    ;; ...ok, now we can call 'this'.
    
    ;; We have to initialize the scrollbars
    (send this init-manual-scrollbars #f #f 1 1 0 0)
    
    ;; We then need to call on-size so that the bitmap and
    ;; drawing context are initialised
    (on-size (send (send this get-parent) get-width) (send (send this get-parent) get-height))))

)


