;; ##################################################################################
;; # ============================================================================== #
;; # MrEd Designer - project-data.ss                                                #
;; # http://mreddesigner.lozi.org                                                   #
;; # Copyright (C) Lozi Jean-Pierre, 2004 - mailto:jean-pierre@lozi.org             #
;; # ============================================================================== #
;; #                                                                                #
;; # This program is free software; you can redistribute it and/or                  #
;; # modify it under the terms of the GNU General Public License                    #
;; # as published by the Free Software Foundation; either version 2                 #
;; # of the License, or (at your option) any later version.                         #
;; #                                                                                #
;; # This program is distributed in the hope that it will be useful,                #
;; # but WITHOUT ANY WARRANTY; without even the implied warranty of                 #
;; # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                  #
;; # GNU General Public License for more details.                                   #
;; #                                                                                #
;; # You should have received a copy of the GNU General Public License              #
;; # along with this program; if not, write to the Free Software                    #
;; # Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.    #
;; #                                                                                #
;; ##################################################################################

;; This variable contains all current project information
(define project-data-global '((project (label "Untitled Project")
                                       (path "")
                                       (selected #f)
                                       (collapsed #f))))

;; This boolean is true iff no change have been made since last save
(define no-change-since-last-save #t)

;; This functions allows to add a child to the project
(define (project-data-add-child child)
  (let
    ((add-child (lambda (children)
                  (if (eq? children 'no-prop)
                    (list child)
                    (append children (list child))
                  ))))
    (set! project-data-global (list (set-property-value (car project-data-global)
                                                        'children
                                                        add-child)))
  )
)

;; This function adds and element whose parent is determined by parent-id
(define (project-data-add-element element parent-id)
  (let*
    ((add-child (lambda (children)
                  (if (eq? children 'no-prop)
                    (list element)
                    (append children (list element))
                  )))
     (cons-line (lambda (line)
                  (if (and (assq 'id (cdr line))
                           (eq? (cadr (assq 'id (cdr line))) parent-id))
                    (set-property-value line 'children add-child)
                    line))))
    (set! project-data-global (apply-function-on-data-list project-data-global cons-line)))
)

;; This function returns the element's line, given its id
(define (project-data-get-element-by-id id)
  ;; We don't know the output yet
  (let ((output '?))
    ;; This function is applied to each line
    (define (helper element)
      (if (and (not (eq? (car element) 'project))
             (eq? (get-property-value element 'id) id))
          ;; Is it the right line? Save it then. 
          (set! output element)))
    ;; Apply our procedure on each element
    (apply-procedure-on-data-list project-data-global helper)
    ;; Return the output
    output))


;; This function returns the element parent's line, or #f.
(define (project-data-get-parent element)
  ;; Get the parent line...
  (let ((temp-output (project-data-get-element-by-id (get-property-value element 'parent))))
    ;; Is it valid?
    (if (eq? temp-output '?) #f temp-output)))


;; This function updates the project path and label
(define (project-data-set-path-label path)
  (let-values 
    (((base name must-be-dir?)(split-path path)))
    (let
      ((label (car (regexp-match "[^.]+" (path->string name)))))
      (set! project-data-global (list (set-property-value (car project-data-global)
                                                          'label
                                                          (lambda (x) label))))
      (set! project-data-global (list (set-property-value (car project-data-global)
                                                          'path
                                                          (lambda (x) path)))))))


;; This function returns the project path
(define (project-data-get-path)
  (get-property-value (car project-data-global) 'path))


;; This very useful function does not have much to do with the project data, but
;; I had to find a place to declare it :)
(define (symbol-append . L)
    (string->symbol (apply string-append (map symbol->string L))))

