;; ##################################################################################
;; # ============================================================================== #
;; # MrEd Designer - properties-widget.ss                                           #
;; # http://mreddesigner.lozi.org                                                   #
;; # Copyright (C) Lozi Jean-Pierre, 2004 - mailto:jean-pierre@lozi.org             #
;; # ============================================================================== #
;; #                                                                                #
;; # This program is free software; you can redistribute it and/or                  #
;; # modify it under the terms of the GNU General Public License                    #
;; # as published by the Free Software Foundation; either version 2                 #
;; # of the License, or (at your option) any later version.                         #
;; #                                                                                #
;; # This program is distributed in the hope that it will be useful,                #
;; # but WITHOUT ANY WARRANTY; without even the implied warranty of                 #
;; # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                  #
;; # GNU General Public License for more details.                                   #
;; #                                                                                #
;; # You should have received a copy of the GNU General Public License              #
;; # along with this program; if not, write to the Free Software                    #
;; # Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.    #
;; #                                                                                #
;; ##################################################################################

;; The properties-widget file :
;; We derive a canvas, and then code the whole widget from scratch :)

;; We declare the class...
(define properties-widget%
  ;; We derive it...
  (class canvas%
    ;; Constants
    (define k-left-margin 0)
    (define k-line-height 15)
    (define k-field-height 14)
    (define k-first-column-width 110)
    (define k-empty-height 288)
    ;; Those constants are used for drawing
    (define k-white              (make-object color% 255 255 255))
    (define k-black              (make-object color% 0 0 0))
    (define k-dkgray             (make-object color% 168 166 155))
    (define k-white-brush        (make-object brush%  k-white 'solid))
    (define k-active-brush       (make-object brush%  k-black 'hilite))
    (define k-gray-brush         (make-object brush% (get-panel-background) 'solid))
    (define k-bold-font          (make-object font% 11 'system 'normal 'bold #f 'smoothed #t))
    (define k-invisible-pen      (make-object pen%  (make-object color% 0 0 0) 0 'transparent))
    (define k-xor-brush          (make-object brush% (make-object color% 0 0 0) 'xor))
    ;; Under MacOS, 'modern just does not appear to work! Therefore we have to define a special font
    ;; under this OS (Courier).
    (define k-edit-font          (if (or (eq? (system-type) 'macosx) (eq? (system-type) 'macos))
                                     (make-object font% 12 "Courier" 'modern 'normal 'normal #f 'unsmoothed #t)
                                     (make-object font% 12 'modern 'normal 'normal #f 'unsmoothed #t)))
    ;; Windows doesn't highlight smoothed text in a nice way.
    (define k-normal-font        (if (eq? (system-type) 'windows)
                                   (make-object font% 11 'system 'normal 'normal #f 'unsmoothed #t)
                                   (make-object font% 11 'system 'normal 'normal #f 'smoothed #t)))
    
    ;; The edit font vertical offset (currently set to 0, except for MacOS).
    (define k-font-y-offset      (if (or (eq? (system-type) 'macosx) (eq? (system-type) 'macos)) 3 0))
    
    ;; Fields
    (define data '())
    (define cursor-pos 0)
    (define select-pos 0)
    (define y-offset 0)
    (define current-field "")
    (define id 0)
    (define dragging #f)
    (define motion-timestamp 0)
    (define default-values '?)
    ;; Why using a bitmap and a drawing context as a class field instead
    ;; of directly declare it in the on-paint body? The matter is,
    ;; that if we did so a bitmap would be instantiated for each call of
    ;; the on-paint method. After testing it, even on a pretty powerful
    ;; pc, if the on-paint method is too often called, the program runs out
    ;; of resources and returns an error when instantiating a bitmap.
    (define bitmap #f)
    (define dc #f)
    
    ;; Public methods
    (public set-data 
            empty
            update-in
            update-buttons
            edit-current-value-with-dialog
            reset-current-value
            reset-widget)
    
    ;; The 'data' modifier
    (define (set-data new-data) 
      (set! data new-data)
      ;; We need to update the window height consequently...
      (send
       properties-frame
       resize 
       (max (send properties-frame get-width) 350) 
       (if (null? data) k-empty-height (+ 2 (max-height) vertical-window-offset (send properties-horizontal-panel get-height))))
      (on-paint))
    
    ;; Empties the widget
    (define (empty)
      (set-data '()))
    
    ;; Access method for the currently selected field
    (define (get-selected-field)
      (define (iter L)
        (cond ((null? L) #f)
              ((< (length (car L)) 3) (iter (cdr L)))
              ((eq? (caddar L) #t) (car L))
              (else (iter (cdr L)))))
      (iter data))
    
    ;; Acces method for the 'name' field
    (define (get-field name)
      (define (iter L)
        (cond ((null? L) #f)
              ((eq? (caar L) name) (car L))
              (else (iter (cdr L)))))
      (iter data))
    
    ;; Value modifier for the 'name' field
    (define (set-field-value name value)
      (define (aux L)
        (cond ((null? L) L)
              ((eq? (caar L) name) (cons (list (caar L) value (caddar L)) (aux (cdr L))))
              (else (cons (car L) (aux (cdr L))))))
      (set! data (aux data)))
    
    ;; Value modifier for the field number num
    (define (set-field-value-num num value)
      (define (aux L incr)
        (cond ((null? L) L)
              ((eq? incr num) (cons (list (caar L) value (caddar L)) (aux (cdr L) (add1 num))))
              (else (cons (car L) (aux (cdr L) (sub1 num))))))
      (set! data (aux data 0)))
    
    ;; Selects the 'name' field
    (define (select-field name)
      (define (aux L)
        (cond ((null? L) L)
              ((< (length (car L)) 3) (cons (car L) (aux (cdr L))))
              ((eq? (caar L) name) (cons (list (caar L) (cadar L) #t) (aux (cdr L))))
              (else (cons (list (caar L) (cadar L) #f) (aux (cdr L))))))
      (set! data (aux data)))
    
    ;; Selects next selectable field
    (define (select-next-field)
      (define (aux L sel i)
        (cond ((null? L) L)
              ((eq? sel #t) (cons (list (caar L) (cadar L) #t) (aux (cdr L) #f (+ i 1))))
              ((< (length (car L)) 3) (cons (car L) (aux (cdr L) sel (+ i 1))))
              ((and (eq? (caddar L) #t) (not (>= i (length L)))) (cons (list (caar L) (cadar L) #f) (aux (cdr L) #t (+ i 1))))
              (else (cons (list (caar L) (cadar L) #f) (aux (cdr L) #f (+ i 1))))))
      (set! data (aux data #f 0)))
    
    ;; Returns the number of lines
    (define (get-number-of-lines)
      (define (aux L)
        (if (null? L)
            0
            (+ 1 (aux (cdr L)))))
      (aux data))
    
    ;; Returns active line number
    (define (active-line-num)
      (define (iter L incr)
        (cond ((null? L) incr)
              ((< (length (car L)) 3) (iter (cdr L) (+ incr 1)))
              ((caddar L) incr)
              (else (iter (cdr L) (+ incr 1)))))
      (iter data 0))
    
    ;; Returns the field, given its number
    (define (get-field-from-num num)
      (define (iter L incr)
        (cond ((null? L) #f)
              ((zero? incr) (car L))
              (else (iter (cdr L) (- incr 1)))))
      (iter data num))
    
    ;; Returns #t if the y-line is a title line
    (define (is-title-line? y)
      (define (iter L y)
        (cond ((null? L) #f)
              ((= y 0) (= 2 (length (car L))))
              (else (iter (cdr L) (sub1 y)))))
      (iter data y))
    
    ;; Returns #t if the line is selectable
    (define (is-selectable-line? y)
      (define (iter L y)
        (cond ((null? L) #f)
              ((= y 0) (and (not (= 2 (length (car L))))
                            (not (eq? (caar L) 'parent))))
              (else (iter (cdr L) (sub1 y)))))
      (iter data y))
    
    ;; Data->string converter
    (define (to-string d)
      (cond ;((eq? 'na d) "")
            ((string? d) d)
            ((null? d) "()")
            ((number? d) (number->string d))
            ((symbol? d) (symbol->string d))
            ((and (boolean? d) d) "#t")
            ((boolean? d) "#f")
            ((pair? d)(format "~s" d))
            (else "")))
    
    ;; This function updates the scroll bars properties...
    (define (update-scroll-bars)
      ;; We need to get the virtual size...
      (let-values (((width height) (send this get-virtual-size)))
        ;; Then, we update the scroll range for vertical scroll bar.
        (if (null? data)
            ;; If the data-list is empty, the scroll-bar is disabled.
            (send this set-scroll-range 'vertical 0)
            ;; The right range is the maximal height minus the virtual height.
            (send this set-scroll-range 'vertical (max 0 (- (max-height) height))))
        ;; If the window can contain the whole height, no y offset is needed (not really
        ;; needed for this widget, but sometimes needed under GNU/Unix systems).
        (if (= 0 (send this get-scroll-range 'vertical)) (set! y-offset 0) (on-paint))
        ;; This is the number of scroll steps - vertically we want to scroll not exactly one
        ;; page, but one page minus one line (as it is always implemented).
        (send this set-scroll-page 'vertical (max 1 (- height k-line-height)))))
    
    ;; This function updates the drawing bitmap...
    (define (update-bitmap width height)
      ;; Is the size valid?
      (if (and (> width 0) (> height 0))
          ;; We update the bitmap size...
          (set! bitmap (make-object bitmap% width (max-height) #f))
          (set! bitmap (make-object bitmap% 1 (max-height) #f)))
      ;; And instantiate the associated dc.
      (set! dc (instantiate bitmap-dc% (bitmap))))
    
    ;; Returns true iff the string is valid for the property
    (define (valid-str? str prop)
      ;; The "not available" property has been removed in this version.
      ;(if (string=? str "na") 
      ;    #t
      (case prop
        ;; If the property is a label, str is always valid.
        ((label) #t)
        ;; This property can be either a number or a string depending on the current widget.
        ((init-value) (if (eq? (car (project-data-get-element-by-id id)) 'text-field)
                          #t
                          (and (integer? (string->number str)) (<= 0 (string->number str) 10000))))
        ;; Those properties are positive integers in [0,10000].
        ((width height min-width min-height border spacing) 
         (and (integer? (string->number str)) (<= 0 (string->number str) 10000)))
        ;; Those properties are positive integers in [0,1000].
        ((vert-margin horiz-margin) (and (integer? (string->number str)) (<= 0 (string->number str) 1000)))
        ;; Those properties are integers in [-10000,10000].
        ((x y min-value max-value) (and (integer? (string->number str)) (<= -10000 (string->number str) 10000)))
        ;; This property is an integer in [1,10000].
        ((range) (and (integer? (string->number str)) (<= 1 (string->number str) 10000)))
        ;; These properties are booleans.
        ((enabled stretchable-width stretchable-height) (cond ((string=? str "#t") #t)
                                                              ((string=? str "#f") #t)
                                                              (else #f)))
        ;; This property is an alignment list.
        ((alignment)
         ;; We read the value...
         (let ((read-value (read-from-string str (lambda (err) (void)) (lambda () #f))))
           ;; And perform necessary tests...
           (and read-value
                (pair? read-value)
                (list? read-value)
                (eq? (car read-value) 'quote)
                (not (null? (cdr read-value)))
                (list? (cadr read-value))
                (= (length (cadr read-value)) 2)
                ;; The first element of the quoted list is either left, center or right.
                (member (caadr read-value) '(left center right))
                ;; The second element of the quoted list is either top, center or bottom.
                (member (cadadr read-value) '(top center bottom)))))
        
        ;; This property is a list of strings.
        ((choices)
         ;; We read the value...
         (let ((read-value (read-from-string str (lambda (err) (void)) (lambda () #f))))
           ;; And perform necessary tests...
           (and read-value
                (pair? read-value)
                (list? read-value)
                (eq? (car read-value) 'quote)
                (not (null? (cdr read-value)))
                (list? (cadr read-value))
                ;; We get the non-string values and check that there are none.
                (null? (filter (lambda (elt) (not (string? elt))) (cadr read-value))))))
                
        ;; This property differs depending on the widget it belongs to.
        ((style)
         ;; We read the value...
         (let ((read-value (read-from-string str (lambda (err) (void)) (lambda () #f))))
           ;; And perform necessary tests...
           (and read-value
                (pair? read-value)
                (list? read-value)
                (eq? (car read-value) 'quote)
                (not (null? (cdr read-value)))
                (list? (cadr read-value))
                ;; This function checks that the cadr contains only allowed values...
                (letrec ((is-valid-list? (lambda (allowed-elements)
                                           ;; We get the invalid values and check that there are none.
                                           (null? (filter (lambda (elt) (not (memq elt allowed-elements))) (cadr read-value))))))
                  ;; And then it depends on the widget type...
                  (case (car (project-data-get-element-by-id id))
                    ((button horizontal-panel vertical-panel panel) (is-valid-list? '(border deleted)))
                    ((text-field) (is-valid-list? '(single multiple hscroll password vertical-label horizontal-label deleted)))
                    ((canvas) (is-valid-list? '(border control-border vscroll hscroll gl deleted no-autoclear))) 
                    ((check-box group-box-panel message) (is-valid-list? '(deleted)))
                    ((choice) (is-valid-list? '(vertical-panel horizontal-panel deleted)))
                    ((frame) (is-valid-list? '(no-resize-border no-caption no-system-menu mdi-parent mdi-child toolbar-button hide-menu-bar)))
                    ((gauge radio-box slider) (is-valid-list? '(horizontal vertical vertical-label horizontal-label deleted)))
                    ((list-box) (is-valid-list? '(single multiple extended vertical-label horizontal-label deleted)))
                    ((tab-panel) (is-valid-list? '(no-border deleted))))))))
                  
        ((callback paint-callback)(read-from-string str (lambda (err) (void)) (lambda () #f)))
        (else #t)))
    
    ;; Returns the converted string
    (define (convert-str str prop)
      ;; First, we want to know the widget type...
      (let ((type (car (project-data-get-element-by-id id))))
        ;; If the string is "not available", go and search the default values
        ;(if (string=? str "na")
        ;    (to-string (cadr (assq prop (cdr default-values))))
        ;; Otherwise, it depends on the field type
        (case prop
          ;; A label?
          ((label) (if (member type '(canvas choice gauge list-box radio-box slider text-field)) ;; Those controls can have a #f label...
                       (if (string=? str "#f") #f str)
                       str))
          ;; This property can be either a number or a string depending on the current widget.
          ((init-value) (if (eq? type 'text-field)
                            str
                            (string->number str)))
          ;; A number? string->number will solve this for us.
          ((width height x y min-width min-height
                  border spacing vert-margin horiz-margin min-value max-value
                  range selection) 
           (string->number str))
          ;; A boolean? then we need a if.
          ((enabled stretchable-width stretchable-height) (if (string=? str "#t") #t #f))
          ;; A list?
          ((alignment style choices callback paint-callback) (read-from-string str)))))
    
    ;; Returns the converted list
    (define (convert data-list)
      ;; Converts each element of data-list...
      (define (helper data-list)
        (cond ((null? data-list) data-list)
              ((< (length (car data-list)) 3) (cons (car data-list) (helper (cdr data-list))))
              ((or (eq? (caar data-list) 'parent))
               (cons (list (caar data-list)
                           (cadr (assq (caar data-list) (cdr (get-selected-element project-data-global))))
                           (caar data-list))(helper (cdr data-list))))
              (else
               (cons (list (caar data-list) (convert-str (cadar data-list) (caar data-list)) (caddar data-list)) (helper (cdr data-list))))))
      ;; We call the helper function...
      (helper data-list))
      
    ;; For only one element
    (define (convert-elt elt)
      (car (convert (list elt))))
    
    ;; Updates the data list.
    (define (update-in)
      ;; We need a parser in order to build our own data
      (letrec ((parse-list (lambda (L)
                             (cond ((null? L) L)
                                   ((eq? (caar L) 'selected)(parse-list (cdr L)))
                                   ((eq? (caar L) 'collapsed)(parse-list (cdr L))) 
                                   ((eq? (caar L) 'children)(parse-list (cdr L)))
                                   ((eq? (caar L) 'pointer)(parse-list (cdr L)))
                                   ((eq? (caar L) 'id)
                                    (set! id (cadar L))
                                    (parse-list (cdr L)))
                                   ((eq? (caar L) 'necessary)(cons (list "Necessary Properties" #f) (parse-list (cdr L))))
                                   ((eq? (caar L) 'optional)(cons (list "Optional Properties" #f) (parse-list (cdr L))))
                                   (else (cons (list (caar L) (to-string (cadar L)) #f) (parse-list (cdr L)))))))
               (selected (get-selected-element project-data-global))
              )
        (if (and hierarchy-widget selected)
            (begin
              (set-data (parse-list (cdr selected)))
              ;; Update the default values
              (set! default-values (eval (string->symbol (string-append (symbol->string (car selected))
                                                                        "-default-values"))))))))
    
    ;; Updates the external environment. L contains either nothing or a boolean
    ;; specifying whether the project has to be reloaded.
    (define (update-out last-value-changed . L)
      ;; This subfunction updates the line...
      (define (update-line data-list)
        (if (and
             ;; Do the line have an id?
             (assq 'id (cdr data-list))
             ;; Is this the right line?
             (eq? (cadr (assq 'id (cdr data-list))) id))
              ;; Then, we build the line!
              (let ((line-output (append
                                  (list (car data-list))
                                  (if (assq 'children (cdr data-list))
                                      (list (assq 'children (cdr data-list)))
                                      '())
                                  (list (assq 'pointer (cdr data-list)))
                                  (list (assq 'selected (cdr data-list)))
                                  (list (assq 'collapsed (cdr data-list)))
                                  (list (list 'id id))
                                  ;; Let's develop a little filter
                                  (letrec ((f (lambda (L)
                                                (cond ((null? L) L)
                                                      ((and (string? (caar L))
                                                            (string=? "Necessary Properties" (caar L))) (cons '(necessary) (f (cdr L))))
                                                      ((and (string? (caar L))
                                                            (string=? "Optional Properties" (caar L))) (cons '(optional) (f (cdr L))))
                                                      ;; We *don't* want the third element!
                                                      (else (cons (list (caar L) (cadar L)) (f (cdr L))))))))
                                    ;; ...and apply it.
                                    (f (convert data))))))
                ;; We update the preview... which computes the line once again...
                (preview-update-widget line-output convert-elt last-value-changed))
            ;; Otherwise, we don't modify the element.
            data-list))
        ;; The main function body...
        (set! project-data-global (apply-function-on-data-list project-data-global update-line))
        
        ;; Do we need to reload the project?
        (if (and (not (null? L)) (eq? (car L) #t))
            ;; Yes - then reload everything.
            (preview-update-preview project-data-global project-data-global))
        
      ;; We update the hierarchy frame...
      (send hierarchy-widget set-data project-data-global))
    
    ;; Returns the maximum height of the control
    (define (max-height)
      (let ((line-num (get-number-of-lines)))
        ;; Is the control empty?
        (if (= 0 line-num)
            ;; Therefore, return the default empty size...
            k-empty-height
            ;; Otherwise we return the *real* maximal height.
            (* k-line-height (get-number-of-lines)))))
    
    ;; This function updates the toolbox frame buttons
    (define (update-buttons)
      (if (get-selected-field)
          (begin
            (send properties-edit-button enable #t)
            (send properties-rvalue-button enable #t))
          (begin
            (send properties-edit-button enable #f)
            (send properties-rvalue-button enable #f)))
      (if (null? data)
          (send properties-rwidget-button enable #f)
          (send properties-rwidget-button enable #t)))
    
    ;; This dialog box is shown when the user changes the style property for either a frame, a horizontal-panel,
    ;; a vertical-panel, a tab-panel or a group-box-panel.
    (define (show-reload?-dialog-box)
      (message-box/custom "Warning"
        (string-append "You are about to change the style and/or choice property of a container. The preview will not reflect this change"
                       " until you reload the project (i.e. choose 'Reload the project', or close and reopen the project). "
                       "What do you want to do?")
        "Reload the project" "Don't update the preview" "Cancel" properties-frame '(default=1 number-order caution) 3))
    
    ;; This dialog box is shown when the user tries to set an invalid value to a property.
    (define (show-invalid-dialog-box field property)
      (message-box "Error" (string-append "The string '"
                                          field
                                          "' is not valid for the '"
                                          property
                                          "' property.") properties-frame '(ok stop)))
    
    ;; This dialog box is shown when the user tries to change a callback.
    (define (show-callback-warning-box)
          (message-box "Warning" 
                       (string-append "Please notice that modifying a callback or a paint-callback might cause errors and "
                                                "generate a bad code if the function contains errors. Therefore, you should check"
                                                " the callback twice before validing it. Are you sure you want to continue?") 
                       properties-frame '(yes-no caution)))
    
    ;; A function that performs the necessary updates after the user 
    ;; set the selected field's value to new-field-value
    (define (perform-updates new-field-value)
      ;; A simple function that performs the necessary updates after having changed
      ;; a propertie's value.
      (define (helper result reload-widget)
        ;; We update the current field...
        (set! current-field (to-string result))
        ;; The data...
        (set-field-value (car (get-selected-field)) current-field)
        ;; The cursor and selection positions...
        (set! cursor-pos (string-length (to-string (cadr (get-selected-field)))))
        (set! select-pos cursor-pos)
        ;; The focus...
        (send this focus)
        ;; The no-change-since-last-change boolean...
        (set! no-change-since-last-save #f)
        ;; We update the external context...
        (if reload-widget
            (update-out (car (get-selected-field)) reload-widget)
            (update-out (car (get-selected-field))))
        ;; And then repaint the widget
        (on-paint))
      
      ;; If the current field's value is the same as the new field's value, we have nothing to do.
      (if (not (string=? current-field (to-string new-field-value)))
          ;; Can we update the preview for the given property?
          (if (and (container? (project-data-get-element-by-id id)) 
                   (memq (car (get-selected-field)) '(style choices)))
              (let ((reload-result (show-reload?-dialog-box)))
                (cond ((= 1 reload-result) 
                       ;; "Don't update the preview" has been selected, then just update everything
                       ;;  telling update-out that we want to reload the preview.
                       (helper new-field-value  #t))
                      ((= 2 reload-result)
                       ;; "Don't update the preview" has been selected, then just update everything
                       ;;  telling update-out that we do want to update the preview.
                       ;; ...telling update-out that we do not want to reload the preview.
                       (helper new-field-value  #f))))
              ;; We can update the preview - just update everything...
              (if (memq (car (get-selected-field)) '(callback paint-callback))
                  (if (eq? 'yes (show-callback-warning-box))
                      (helper new-field-value #f))
                  (helper new-field-value #f)))))
    
    ;; This function opens a dialog box allowing to edit
    ;; the current element value
    (define (edit-current-value-with-dialog)
      (let* ((selected-field (get-selected-field))
             (message (string-append "Please enter the new value for the " (symbol->string (car selected-field)) " property."))
             (result (get-text-from-user "Value edition..." message properties-frame current-field '())))
        (if result
            (if (not (valid-str? result (car selected-field)))
                (show-invalid-dialog-box result (symbol->string (car selected-field)))
                ;; Otherwise, save it.
                (perform-updates result)))))
    
    ;; This function resets the currently selected value
    (define (reset-current-value)
      ;; Here is the main loop...
      (define (iter L selected-field)
        ;; If the list is empty, the work is done...
        (cond ((null? L) (void))
              ;; If it is not a real field, continue with the rest of the list...
              ((< (length (car L)) 2) (iter (cdr L) selected-field))
              ;; Is the the right field?
              ((eq? (car selected-field) (caar L))
               ;; Then reset it, and perform necessary updates...
               (perform-updates (cadar L)))
              ;; If this is not the right, field, then parse the remaining list...
              (else (iter (cdr L) selected-field))))
      
      ;; Call the main loop...
      (iter (cdr default-values) (get-selected-field)))
       
    ;; This function is called when the user validates a changed property either with the mouse
    ;; or with the keyboard.
    (define (update-selected-field)
      (let ((selected-field (get-selected-field)))
        (if (not (valid-str? current-field (car selected-field)))
            ;; If the value is not valid, display a message
            (show-invalid-dialog-box current-field (symbol->string (car selected-field)))
            ;; Otherwise, save it
            (if (not (string=? (cadr selected-field) current-field))
                ;; Is the current element a container with a 'style' property?
                (if (and (container? (project-data-get-element-by-id id))
                         (memq (car selected-field) '(style choices)))
                    ;; Then ask the user whether he wants to reload the project...
                    (let ((reload-result (show-reload?-dialog-box)))
                      (cond ((= 1 reload-result)
                             (set-field-value (car selected-field) current-field)
                             (update-out (car selected-field) #t))
                            ((= 2 reload-result)
                             (set-field-value (car selected-field) current-field)
                             (update-out (car selected-field)))))
                    ;; Otherwise, just update the field and call update-out without extra argument...
                    (if (memq (car selected-field) '(callback paint-callback))
                         (if (eq? 'yes (show-callback-warning-box))
                             (set-field-value (car selected-field) current-field)
                             (update-out (car selected-field)))
                         (begin
                           (set-field-value (car selected-field) current-field)
                           (update-out (car selected-field)))))))))
    
    ;; This function resets the whole widget
    (define (reset-widget)
      ;; The main loop...
      (define (helper L num need-update-out)
        ;; If the list is empty, the work is done!
        (cond ((null? L) (void))
              ;; The first element is not a field or is un selectable - ignore it.
              ((or (< (length (car L)) 3)
                   (not (is-selectable-line? num))) (helper (cdr L) (add1 num) need-update-out))
              ;; Otherwise it is a field - call reset-field then!
              (else (reset-field (caar L) need-update-out)
                    (helper (cdr L) (add1 num) need-update-out))))
      ;; First, we need to have a confirmation from the user...
      (if (eq? 'yes (message-box "Warning" 
                                 "Are you sure you want to reset the whole widget?" #f 
                                 '(yes-no caution)))
          ;; If this is a container with a 'style' property?
          (if (and (container? (project-data-get-element-by-id id))
                   (or (and (get-field 'style)
                            (not (string=? (cadr (get-field 'style)) (to-string (cadr (assq 'style (cdr default-values)))))))
                       (and (get-field 'choices)
                            (not (string=? (cadr (get-field 'choices)) (to-string (cadr (assq 'choices (cdr default-values)))))))))
              ;; Then ask the user whether he wants to reload the preview...
              (let ((reload-result (show-reload?-dialog-box)))
                (cond ((= 1 reload-result) (helper data 0 'no-update-out) (update-out (car (get-selected-field)) #t))
                      ((= 2 reload-result) (helper data 0 'update-out))))
              ;; Otherwise just call the main loop...
              (helper data 0 'update-out))))
    
    ;; This function resets the field named "field"
    (define (reset-field field need-update-out)
      ;; Here is the main loop...
      (define (helper L field)
        ;; If the list is empty, the field has not been found - no error needed.
        (cond ((null? L) (void))
              ;; The first element is not a field - ignore it.
              ((< (length (car L)) 2) (helper (cdr L) field))
              ;; Is it the right field?
              ((eq? field (caar L))
               ;; Then, update its value...
               (set-field-value field (to-string (cadar L))))
              ;; Else, continue with the remaining list...
              (else (helper (cdr L) field))))
      
      ;; We call the main loop...
      (helper (cdr default-values) field)
      ;; And then perform necessary updates...
      (send this focus)
      (set! no-change-since-last-save #f)
      (if need-update-out
          (update-out field))
      (on-paint))
    
    ;; Overriden methods
    (override on-scroll on-size on-paint on-event on-char)

    ;; +----------------------------------------------------------------------+
    ;; | Function properties-widget::on-scroll                                 |   
    ;; +----------------------------------------------------------------------+
    (define (on-scroll scroll-event)

         ;; Then it depends on the event type...
         (let ((type (send scroll-event get-event-type)))
           ;;
           ;; Notice : I am forced to use eq?, because I have noticed a bug under MacOS X : if the event is a thumb event,
           ;; DrScheme crashes if we try to evaluate the result given by get-event-type! It looks to return an invalid
           ;; pointer, at a system adress, maybe 0. Therefore I just compare pointers.
           ;;
           ;; If the event type is line down...
           (cond ((eq? type 'line-down)
                  ;; ...then we need the virtual size...
                  (let-values (((width height) (send this get-virtual-size)))
                    ;; ...update the scroll-bars positions (to force scrolling more than 1 step [1 line, actually])...
                    (send this set-scroll-pos 'vertical (min (+ (send this get-scroll-pos 'vertical) k-line-height -1) 
                                                             (send this get-scroll-range 'vertical)))
                    ;; ...and the vertical offset.
                    (set! y-offset (min (+ y-offset k-line-height) (send this get-scroll-range 'vertical)))))
                 
                 ;; If the event type is line-up...
                 ((eq? type 'line-up)
                  ;; ...then we update the scroll-bars positions (to force scrolling more than 1 step...)...
                  (send this set-scroll-pos 'vertical (max (- (send this get-scroll-pos 'vertical) k-line-height -1) 0))
                  ;; ...and the vertical offset.
                  (set! y-offset (max (- y-offset k-line-height) 0)))
                 
                 ;; Otherwise, we do not have to update the scroll-bars positions, just to update the vertical offset...
                 (else (set! y-offset (send scroll-event get-position)))))
        

      
      ;; And then, we directly draw the bitmap, which has already been drawed during the last
      ;; call of (on-paint), at the right position - which is *much* faster than calling on-paint
      ;; directly and allows a smooth scrolling.
      (send (send this get-dc) draw-bitmap bitmap 0 (- y-offset)))
    
    ;; +----------------------------------------------------------------------+
    ;; | Function properties-widget::on-size                                  |   
    ;; +----------------------------------------------------------------------+
    (define (on-size width height)
      (if (not (null? data))
          ;; Is the height is too important...
          (if (> height (max-height))
              ;; Set it to the maximal height.
              (send properties-frame
                    resize 
                    (send properties-frame get-width)
                    (+ 2 ;; (due to the borders)
                       (max-height) vertical-window-offset
                       (send properties-horizontal-panel get-height)))))
      ;; We update the drawing bitmap...
      (update-bitmap width height)
      ;; ...and the scroll bars.
      (update-scroll-bars))
    
    ;; +----------------------------------------------------------------------+
    ;; | Function properties-widget::on-paint                                 |   
    ;; +----------------------------------------------------------------------+
    (define (on-paint)
      (let* ((width              (send this get-width))
             (height             (send this get-height)))
        
        ;; Data list parser
        (define (parse-data-list data-list incr)
          (if (not (null? data-list))
              ;; Is it a caption or a real line?
              (cond ((string? (caar data-list)) (begin
                                                  (send dc set-text-foreground k-black)
                                                  (send dc set-font k-bold-font)
                                                  (send dc draw-text (caar data-list) (+ k-left-margin 4) (+ (* incr k-line-height) 1))
                                                  (parse-data-list (cdr data-list) (add1 incr))))
                    (else 
                     ;; Black text on white background...
                     (send dc set-brush k-white-brush)
                     (send dc set-text-foreground k-black)
                     
                     ;; Draw the line's two white rectangles...
                     (send dc draw-rectangle k-left-margin (+ (* incr k-line-height) 1) k-first-column-width k-field-height)
                     (send dc
                           draw-rectangle 
                           (+ k-first-column-width k-left-margin 1)
                           (+ (* incr k-line-height) 1)
                           (- width (+ k-first-column-width k-left-margin 2))
                           k-field-height)
                     
                     ;; We choose the right font color...
                     (if (not (is-selectable-line? incr))
                         (send dc set-text-foreground k-dkgray)
                         (send dc set-text-foreground k-black))
                                        
                     ;; Then we chose the right font...
                     (send dc set-font k-normal-font)
                     
                     ;; ...in order to display the property name.
                     (send dc draw-text (symbol->string (caar data-list)) (+ k-left-margin 4) (+ (* incr k-line-height) 1))
                     
                     ;; We display its value
                     (send dc set-font k-edit-font)
                     
                     (send dc draw-text 
                           ;; Do the item have focus?
                           (if (caddar data-list)
                               current-field
                               (cadar data-list))
                           (+ k-left-margin k-first-column-width 4)
                           (+ (* incr k-line-height) k-font-y-offset))
                     
                     ;; Do the item have focus?
                     (if (caddar data-list)
                         (begin
                           ;; We chose the platform-independant 'Active' brush.
                           (send dc set-brush k-active-brush)
                           (send dc
                                 draw-rectangle 
                                 (+ k-first-column-width k-left-margin 4 (* k-char-width (min cursor-pos select-pos)))
                                 (+ (* incr k-line-height) 1)
                                 (* k-char-width (abs (- cursor-pos select-pos)))
                                 k-field-height)
                           
                           (send dc draw-rectangle k-left-margin (+ (* incr k-line-height) 1) k-first-column-width k-field-height)))
                     
                     (parse-data-list (cdr data-list) (add1 incr))))))
              

        ;; We need to clear the drawing context as the same bitmap is used for each on-paint call
        (send dc clear)
        
        ;; Invisible pen
        (send dc set-pen k-invisible-pen)
        
        ;; Gray brush
        (send dc set-brush k-gray-brush)
        
        ;; We draw the background...
        (send dc draw-rectangle 0 0 width (send bitmap get-height))
        
        ;; ...and then it depends...
        (if (null? data)
            ;; ...either the data-list is empty and therefore we draw the right
            ;; warning message...
            (begin
              (send dc set-font k-bold-font)
              (send dc draw-text "<No Selected Element>" (- (/ width 2) 70) (- (/ height 2) 10)))
            ;; ...or we display the list.
            (parse-data-list data 0))
        
        ;; We drawed the whole output into a bitmap in order to avoid flickering :)
        ;; Then we draw the bitmap
        (if (send bitmap ok?)
            (send (send this get-dc) draw-bitmap bitmap 0 (- y-offset)))))
 
    ;; +----------------------------------------------------------------------+
    ;; | Function properties-widget::on-event                                 |   
    ;; +----------------------------------------------------------------------+
    (define (on-event event)
      ;; Takes a local y coordinate and returns the line number
      (define (locate-y y)
        (define (iter y line)
          (if (and (> y 0) (<= y (+ k-line-height 1)))
              line
              (iter (- y k-line-height) (+ 1 line))))
        (iter y 0))
      
      ;; Takes a local x coordinate and returns the text column number
      (define (locate-x x)
        (let ((column (round (/ (- x (+ k-left-margin k-first-column-width 4)) k-char-width))))
          (if (< column 0)
              #f
              (inexact->exact (min column (string-length current-field))))))
      
      ;; Computes data-list in order to focus the given line and un-focus
      ;; all the others
      (define (focus line data-list)
        (define (sub-focus L bool) ;; Works on a sublist
          (if (> (length L) 2)
              (list (car L) (cadr L) bool)
              L))
        (if (null? data-list)
            data-list
            (if (zero? line)
                (cons (sub-focus (car data-list) #t) (focus (- line 1) (cdr data-list)))
                (cons (sub-focus (car data-list) #f) (focus (- line 1) (cdr data-list))))))
      
      ;; Returns #t if the coordinates are in a text area
      (define (is-in-text-area? x y)
        (and
         (not (< x (+ k-left-margin k-first-column-width)))
         (is-selectable-line? (locate-y y))))
      
      ;; On-event body starts here
      (send this focus)
      (let ((x    (send event get-x))
            (y    (+ (send event get-y) y-offset)))
        (if (and (<= y (* (get-number-of-lines) k-line-height)) ;; y coordinate must be valid
                 (> y 0) (not (null? list)))
            (case (send event get-event-type)
              
              ;; Mouse-down event with the left mouse button
              ((left-down)
               (let* ((line-num (locate-y y))
                      (field (get-field-from-num line-num))
                      (selected-field (get-selected-field)))
                 (if (is-selectable-line? line-num)
                     (begin
                       (if selected-field
                             ;; We will not do anything if the event line is the current line.
                             (if (or (not field) (not (eq? (car field) (car selected-field))))
                                 (update-selected-field)))
                             

                      ;; If the previously selected field was not the one the user clicked in...
                      (if (or (not selected-field)
                              (or (not field) (not (eq? (car field) (car selected-field)))))
                          ;; Then select it...
                          (begin (set! data (focus line-num data))
                                 ;; And update the current-field variable.
                                 (set! current-field (cadr (get-selected-field)))))

                       (set! dragging #t)
                       
                       (if (locate-x x)
                           (begin
                             (set! cursor-pos (locate-x x))
                             (set! select-pos cursor-pos))
                           (begin
                             (set! cursor-pos (string-length (cadr (get-selected-field))))
                             (set! select-pos cursor-pos)))
                       
                       (update-buttons)
                       (on-paint)))))
              
              ;; Mouse-up event with the left mouse button
              ((left-up)
               (set! dragging #f)
               ;; This call to on-paint is needed in order to have a selection
               ;; reaching exactly the caret
               (on-paint))
              
              ;; Mouse-move event
              ((motion)
               (if (is-in-text-area? x y)
                   (send this set-cursor (make-object cursor% 'ibeam))
                   (send this set-cursor (make-object cursor% 'arrow)))
               (if (send event dragging?)
                   (begin
                     (if (locate-x x)
                         (set! cursor-pos (locate-x x))
                         (set! cursor-pos (string-length current-field)))
                     ;; Avoid to call (on-paint) too often
                     (if (> (- (send event get-time-stamp) motion-timestamp) 30)
                         (begin
                           (on-paint)
                           (set! motion-timestamp (send event get-time-stamp)))))))))))
    
    
    
    ;; +----------------------------------------------------------------------+
    ;; | Function properties-widget::on-char                                  |   
    ;; +----------------------------------------------------------------------+
    (define (on-char event)
      ;; Data must not be empty, otherwise we won't answer the event.
      (if (not (null? data))
          (case (send event get-key-code)
            
            ;; Released key
            ((release #\nul)
             (void))
            
            ;; Wheel up
            ((wheel-up)
             (on-scroll (new scroll-event% (event-type 'line-up) (direction 'vertical))))
            
            ;; Wheel down
            ((wheel-down)
             (on-scroll (new scroll-event% (event-type 'line-down) (direction 'vertical))))
            
            ;; Left arrow
            ((left)
             (set! cursor-pos (max (- cursor-pos 1) 0))
             (if (not (send event get-shift-down))
                 (set! select-pos cursor-pos)))
            
            ;; Right arrow
            ((right)
             (set! cursor-pos (min (add1 cursor-pos) (string-length current-field)))
             (if (not (send event get-shift-down))
                 (set! select-pos cursor-pos)))  
            
            ;; Delete
            ((#\backspace)
             (if (> (string-length (cadr (get-selected-field))) 0)
                 (begin
                   (if (= cursor-pos select-pos)
                       ;; No selection
                       (begin
                         (set! current-field
                               (string-append
                                (substring current-field 0 (max (- cursor-pos 1) 0))
                                (substring current-field
                                           cursor-pos
                                           (string-length current-field))))
                         (set! cursor-pos (max (- cursor-pos 1) 0))
                         (set! select-pos cursor-pos))
                       
                       ;; Selection
                       (begin
                         (set! current-field
                               (string-append
                                (substring current-field 0 (min cursor-pos select-pos))
                                (substring current-field
                                           (max cursor-pos select-pos)
                                           (string-length current-field))))
                         (set! cursor-pos (min cursor-pos select-pos))
                         (set! select-pos cursor-pos)))))
             (set! no-change-since-last-save #f))
            
            ((#\return)
              (update-selected-field))
            
            ;; Other key
            (else
             ;; Displayable key?
             (if (char? (send event get-key-code))
                 (begin
                   (set! current-field
                         (string-append
                          (substring current-field 0 (min cursor-pos select-pos))
                          (list->string (list (send event get-key-code)))
                          (substring current-field (max cursor-pos select-pos) (string-length current-field))))
                   
                   (set! cursor-pos (add1 (min cursor-pos select-pos)))
                   (set! select-pos cursor-pos)
                   (set! no-change-since-last-save #f))))))
      (on-paint))
    
    ;; Timer
    (define (on-timer)
      (if (and (send this has-focus?) ;; Caret doesn't blink if we don't have focus...
               (not dragging)         ;; ...or if the mouse is being dragged...
               (not (null? data)))    ;; ...or if there is no data
          (begin
            (send (send this get-dc) set-pen k-invisible-pen) ;; Invisible pen
            ;; We use xor in order to draw/erase the caret with the same function
            (send (send this get-dc) set-brush k-xor-brush)
            (send (send this get-dc) draw-rectangle 
                  (+ k-left-margin k-first-column-width 4 (* k-char-width cursor-pos)) 
                  (+ (* (active-line-num) k-line-height) 1 (- y-offset))
                  1
                  14))))
    
    ;; We start the timer
    (define properties-widget-timer (make-object timer% on-timer 500 #f))
    
    ;; We call the canvas initialisation, passing needed arguments
    (super-new (style '(vscroll border)))
    
    ;; We could not define this constant before, because we needed 'this'.
    (define k-char-width (if (or (eq? (system-type) 'macosx) (eq? (system-type) 'macos))
                             7
                             (begin (send (send this get-dc) set-font k-edit-font)
                                    (send (send this get-dc) get-char-width))))
    
    ;; We can not resize a window giving a client size. So this is a workaround
    ;; I finally found to solve the problem. We substract the real
    ;; window height to the client height and then we only will have
    ;; to had this value to the wanted client size.
    (define vertical-window-offset
      (let-values (((w h) (send properties-frame get-client-size)))
        (- (send properties-frame get-height) h)))

    ;; We initialize the scrollbars - maybe isn't it needed?
    ;; For compatibility reasons, we will not delete the following line.
    (send this init-manual-scrollbars #f #f 1 1 0 0)
    
    ;; We then need to call on-size so that the bitmap and
    ;; drawing context are initialized.
    (on-size (send (send this get-parent) get-width) (send (send this get-parent) get-height))))

